#include "MainWindow.hpp"
#include "Utility.h"

#include <cassert>
#include <cmath>


inline uint8_t getXFromOpcode(uint16_t opcode) { return (opcode & 0x0F00) >> 8; }
inline uint8_t getYFromOpcode(uint16_t opcode) { return (opcode & 0x00F0) >> 4; }
inline uint8_t getByteFromOpcode(uint16_t opcode) { return opcode & 0x00FF; }
inline uint16_t getWordFromOpcode(uint16_t opcode) { return opcode & 0x0FFF; }

int Debugger::DisassemblyItems::getNumItems()
{
	return ExecutionEntries->size();
}

sf::Vector2f Debugger::DisassemblyItems::getSize(int index)
{
	return sf::Vector2f(ExecutionEntries->at(index).string.size() * 12, UI::Default::TextFont->getLineSpacing(10));
}

void Debugger::DisassemblyItems::draw(int index, const sf::Vector2f& position, const sf::Vector2f& size, sf::RenderTarget* renderTarget)
{
	sf::Vector2f itemSize = getSize(index);

	sf::RectangleShape leftMargin(sf::Vector2f(20, itemSize.y));
	leftMargin.setPosition(position);
	leftMargin.setFillColor(UI::Default::TabHeaderHover);
	renderTarget->draw(leftMargin);

	sf::Color itemColor;
	if (index % 2 == 0)
		itemColor = UI::Default::WidgetBackground;
	else
		itemColor = sf::Color(100,100,100, 200);

	sf::RectangleShape itemBackground(sf::Vector2f(size.x - 20, itemSize.y));
	itemBackground.setPosition(position.x + 20, position.y);
	if (ExecutionEntries->at(index).current)
		itemColor.r = 255;	
	itemBackground.setFillColor(itemColor);

	renderTarget->draw(itemBackground);
	
	sf::Text text(ExecutionEntries->at(index).string, *UI::Default::TextFont, 10);
	text.setPosition(position.x + 21, position.y);
	renderTarget->draw(text);
}

Debugger::Debugger() : m_executionOutputLines(50)
{
	DisassemblyItemsCallback.ExecutionEntries = &m_execution;
}

void Debugger::onAttach(Chip8::VirtualMachine* vm)
{
	m_vm = vm;
}

//TODO: we will want to cache the disassembled lines so we don't redo ALL the work every damn step

bool Debugger::onStep(uint16_t opcode)
{
	Chip8::Program* program = m_vm->getProgram();
	/*int linesBefore = std::floor(m_executionOutputLines / 2.f);
	int linesAfter = std::ceil(lines / 2.f);*/
	const Chip8::VirtualMachine::State& vmstate = m_vm->saveState();
	uint16_t currentExecutionPoint = vmstate.PC;

	/**
	TODO: finish writing debug output
	we want it of the form:
	byte byte --- disassembly
	byte byte --- disassembly

	so,
	PC PC+1 -- disassembly
	PC+2 PC+3 -- disassembly
	etc...
	*/
	m_execution.clear();
	int lowerLine = (m_executionOutputLines % 2 == 0) ? -m_executionOutputLines : -m_executionOutputLines + 1;
	int upperLine = (m_executionOutputLines % 2 == 0) ? m_executionOutputLines : m_executionOutputLines - 1;
	for (int i = lowerLine; i < upperLine; i += 2)
	{
		/*std::string outputLine = toStrHex(program->bytecode[currentExecutionPoint - m_executionOutputLines + i]) + 
			toStrHex(program->bytecode[currentExecutionPoint - m_executionOutputLines + 1 + i]);*/
		ExecutionEntry newEntry;
		newEntry.string = toStrHex(currentExecutionPoint + i) + " ";
		//
		uint16_t op = ((vmstate.memory[currentExecutionPoint + i] << 8) & 0xFFFF) + (vmstate.memory[currentExecutionPoint + i + 1] & 0xFF);
		if ((currentExecutionPoint + i) == currentExecutionPoint)
		{
			//assert(opcode == op);
			newEntry.current = true;
			newEntry.string +=  toStrHex(opcode) + " " + disassemble(opcode);
			//newEntry.string += " || " + toStrHex( op );
		} else
		{
			newEntry.current = false;
			newEntry.string += toStrHex( op ) + " " + disassemble(op);
		}
		m_execution.push_back(newEntry);
		//m_window->addLogMessage(outputLine);
	}
	
	for (auto iter = m_breakpoints.begin(); iter != m_breakpoints.end(); ++iter)
	{
		if (currentExecutionPoint == *iter)
			return false;
	}

	return true;
}

void Debugger::resume()
{
	Chip8::VirtualMachine::State& vmstate = m_vm->saveState();
	vmstate.status = Chip8::VirtualMachine::Status::Running;
}

void Debugger::showLines(int surroundingLines)
{
	m_executionOutputLines = surroundingLines;
}

void Debugger::setBreakpoint(uint16_t position)
{
	m_breakpoints.push_back(position);
}

void Debugger::removeBreakpoint(uint16_t position)
{
	auto iter = m_breakpoints.begin();

	while (iter != m_breakpoints.end())
	{
		if (*iter == position)
			iter = m_breakpoints.erase(iter);
		else
			++iter;
	}
}

std::string Debugger::disassemble(uint16_t opcode)
{
	switch ((opcode & 0xF000) >> 12)
	{
		case 0x0:
			switch (opcode & 0x00FF)
			{
				case 0xE0:
					return "clearscreen";
					break;
				case 0xEE:
					return "return";
					break;
				case 0xFB:
					return "scrollright";
					break;
				case 0xFC:
					return "scrollleft";
					break;
				case 0xFD:
					return "exit";
					break;
				case 0xFE:
					return "disableschip";
					break;
				case 0xFF:
					return "enableschip";
					break;
				default:
					if (opcode & 0x00F0 == 0xC0)
						return "scrolldown " + toStr(opcode & 0x000F);
					else
						return "";
						//return "NOT SUPPORTED: call 1802 machine code";
			}

			break;
		case 0x1:
			return "jump " + toStrHex(opcode & 0x0FFF);
			break;
		case 0x2:
			return "call " + toStrHex(opcode & 0x0FFF);
			break;
		case 0x3:
			return "skipifregeq V" + toStrHex(getXFromOpcode(opcode)) + ", " + toStrHex(getByteFromOpcode(opcode));
			break;
		case 0x4:
			return "skipifregnoteq V" + toStrHex(getXFromOpcode(opcode)) + ", " + toStrHex(getByteFromOpcode(opcode));
			break;
		case 0x5:
			return "skipifregeq V" + toStrHex(getXFromOpcode(opcode)) + ", V" + toStrHex(getYFromOpcode(opcode));
			break;
		case 0x6:
			return "load V" + toStrHex(getXFromOpcode(opcode)) + ", " + toStrHex(getByteFromOpcode(opcode));
			break;
		case 0x7:
			return "add V" + toStrHex(getXFromOpcode(opcode)) + ", " + toStrHex(getByteFromOpcode(opcode));
			break;
		case 0x8:
			switch (opcode & 0x000F)
			{
				case 0x0:
					return "load V" + toStr(getXFromOpcode(opcode)) + ", V" + toStr(getYFromOpcode(opcode));
					break;
				case 0x1:
					return "or V" + toStrHex(getXFromOpcode(opcode)) + ", V" + toStrHex(getYFromOpcode(opcode));
					break;
				case 0x2:
					return "and V" + toStrHex(getXFromOpcode(opcode)) + ", V" + toStrHex(getYFromOpcode(opcode));
					break;
				case 0x3:
					return "xor V" + toStrHex(getXFromOpcode(opcode)) + ", V" + toStrHex(getYFromOpcode(opcode));
					break;
				case 0x4:
					return "add V" + toStrHex(getXFromOpcode(opcode)) + ", V" + toStrHex(getYFromOpcode(opcode));
					break;
				case 0x5:
					return "sub V" + toStrHex(getXFromOpcode(opcode)) + ", V" + toStrHex(getYFromOpcode(opcode));
					break;
				case 0x6:
					if (m_vm->saveState().correctChip8Shift)
						return "shiftright V" + toStrHex(getXFromOpcode(opcode)) + ", V" + toStrHex(getYFromOpcode(opcode));
					else
						return "shiftright V" + toStrHex(getXFromOpcode(opcode));
					break;
				case 0x7:
					return "sub V" + toStrHex(getYFromOpcode(opcode)) + ", V" + toStrHex(getXFromOpcode(opcode));
					break;
				case 0xE:
					if (m_vm->saveState().correctChip8Shift)
						return "shiftleft V" + toStrHex(getXFromOpcode(opcode)) + ", V" + toStrHex(getYFromOpcode(opcode));
					else
						return "shiftleft V" + toStrHex(getXFromOpcode(opcode));
					break;
				default:
					return "NOT SUPPORTED: unknown alu opcode";
			}
			break;
		case 0x9:
			return "skipifregnoteq V" + toStrHex(getXFromOpcode(opcode)) + ", V" + toStrHex(getYFromOpcode(opcode));
			break;
		case 0xA:
			return "loadi " + toStrHex(getWordFromOpcode(opcode));
			break;
		case 0xB:
			return "NOT SUPPORTED: jumpv0 " + toStrHex(getWordFromOpcode(opcode)) + ", V0";
			break;
		case 0xC:
			return "rand " + toStr(getByteFromOpcode(opcode));
			break;
		case 0xD:
			return "draw V" + toStrHex(getXFromOpcode(opcode)) + ", V" + toStrHex(getYFromOpcode(opcode)) + ", " + toStr(opcode & 0x000F);
			break;
		case 0xE:
			if (opcode & 0x00FF == 0x9E)
				return "skipifkeydown V" + toStrHex(getXFromOpcode(opcode));
			else if (opcode & 0x00FF == 0xA1)
				return "skipifkeynotdown V" + toStrHex(getXFromOpcode(opcode));
			else
				return "NOT SUPPORTED: invalid keypress instruction";
			break;
		case 0xF:
			switch (opcode & 0x00FF)
			{
				case 0x07:
					return "load V" + toStrHex(getXFromOpcode(opcode)) + ", delaytimer";
					break;
				case 0x0A:
					return "waitforkeypress V" + toStrHex(getXFromOpcode(opcode));
					break;
				case 0x15:
					return "loaddelaytimer V" + toStrHex(getXFromOpcode(opcode));
					break;
				case 0x18:
					return "loadsoundtimer V" + toStrHex(getXFromOpcode(opcode));
					break;
				case 0x1E:
					return "addi V" + toStrHex(getXFromOpcode(opcode));
					break;
				case 0x29:
					return "loadifont5 V" + toStrHex(getXFromOpcode(opcode));
					break;
				case 0x30:
					return "loadifont10 V" + toStrHex(getXFromOpcode(opcode));
					break;
				case 0x33:
					return "storebcd V" + toStrHex(getXFromOpcode(opcode));
					break;
				case 0x55:
					return "storemem V" + toStrHex(getXFromOpcode(opcode));
					break;
				case 0x65:
					return "loadmem V" + toStrHex(getXFromOpcode(opcode));
					break;
				case 0x75:
					return "storehp48 V" + toStrHex(getXFromOpcode(opcode));
					break;
				case 0x85:
					return "loadhp48 V" + toStrHex(getXFromOpcode(opcode));
					break;
				default:
					return "NOT SUPPORTED: F0NN, I = 28bit address";
			}
			break;
	}
}