
#include "Chip8VM.hpp"

#include <cstring>
#include <cmath>
#include <cassert>
#include <ctime>

#include <list>

const unsigned char gChip8Fonts[] =
{
	0xF0, 0x90, 0x90, 0x90, 0xF0,	// values for 0
	0x60, 0xE0, 0x60, 0x60, 0xF0,	// values for 1
	0x60, 0x90, 0x20, 0x40, 0xF0,	// values for 2
	0xF0, 0x10, 0xF0, 0x10, 0xF0,	// values for 3
	0x90, 0x90, 0xF0, 0x10, 0x10,	// values for 4
	0xF0, 0x80, 0x60, 0x10, 0xE0,   // values for 5
	0xF0, 0x80, 0xF0, 0x90, 0xF0,   // values for 6
	0xF0, 0x10, 0x10, 0x10, 0x10,	// values for 7
	0xF0, 0x90, 0xF0, 0x90, 0xF0,	// values for 8
	0xF0, 0x90, 0xF0, 0x10, 0x10,	// values for 9
	0x60, 0x90, 0xF0, 0x90, 0x90,	// values for A
	0xE0, 0x90, 0xE0, 0x90, 0xE0,	// values for B
	0x70, 0x80, 0x80, 0x80, 0x70,	// values for C
	0xE0, 0x90, 0x90, 0x90, 0xE0, 	// values for D
	0xF0, 0x80, 0xF0, 0x80, 0xF0,	// values for E
	0xF0, 0x80, 0xF0, 0x80, 0x80	// values for F
};

const unsigned char gSChipFonts[] =
{
	0x00, 0x3C, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x3C, 0x00, //0
	0x00, 0x08, 0x38, 0x08, 0x08, 0x08, 0x08, 0x08, 0x3E, 0x00, //1
	0x00, 0x38, 0x44, 0x04, 0x08, 0x10, 0x20, 0x44, 0x7C, 0x00, //2
	0x00, 0x38, 0x44, 0x04, 0x18, 0x04, 0x04, 0x44, 0x38, 0x00, //3
	0x00, 0x0C, 0x14, 0x24, 0x24, 0x7E, 0x04, 0x04, 0x0E, 0x00, //4
	0x00, 0x3E, 0x20, 0x20, 0x3C, 0x02, 0x02, 0x42, 0x3C, 0x00, //5
	0x00, 0x0E, 0x10, 0x20, 0x3C, 0x22, 0x22, 0x22, 0x1C, 0x00, //6
	0x00, 0x7E, 0x42, 0x02, 0x04, 0x04, 0x08, 0x08, 0x08, 0x00, //7
	0x00, 0x3C, 0x42, 0x42, 0x3C, 0x42, 0x42, 0x42, 0x3C, 0x00, //8
	0x00, 0x3C, 0x42, 0x42, 0x42, 0x3E, 0x02, 0x04, 0x78, 0x00, //9
	0x00, 0x18, 0x08, 0x14, 0x14, 0x14, 0x1C, 0x22, 0x77, 0x00, //A
	0x00, 0x7C, 0x22, 0x22, 0x3C, 0x22, 0x22, 0x22, 0x7C, 0x00, //B
	0x00, 0x1E, 0x22, 0x40, 0x40, 0x40, 0x40, 0x22, 0x1C, 0x00, //C
	0x00, 0x78, 0x24, 0x22, 0x22, 0x22, 0x22, 0x24, 0x78, 0x00, //D
	0x00, 0x7E, 0x22, 0x28, 0x38, 0x28, 0x20, 0x22, 0x7E, 0x00, //E
	0x00, 0x7E, 0x22, 0x28, 0x38, 0x28, 0x20, 0x20, 0x70, 0x00  //F
};

namespace CycleTiming {

	const uint32_t DefaultVMFrequency = 1000;
	//default timing table, every instruction takes 1 cycle
	const uint32_t Default[] = {1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1};

	/**
	COSMAC_VIP, one of the earliest hardware platforms to support Chip8

	Execution times for the fetch and decode loop are 40 machine cycles (181.6 microseconds) for group 0 instructions
	and 68 machine cycles (308.72 microseconds) for all other Chip-8 instructions. Note that these are the execution
	times for the fetch and decode loop only  the execution time for the called routine needs to be added to this
	to get the total execution time for the instruction.

	On COSMAC_VIP, each cycle takes 4.5 us (microseconds)
	This would seem to indicate that the Chip8 interpreter on the system ran effectively at 222kHz, in contrast to the systems
	speed of 1.76MHz.
	*/
	const uint32_t COSMAC_VIP[] = {
		33
	};
} /* end namespace CycleTiming */


inline uint8_t getXFromOpcode(uint16_t opcode) { return (opcode & 0x0F00) >> 8; }
inline uint8_t getYFromOpcode(uint16_t opcode) { return (opcode & 0x00F0) >> 4; }
inline uint8_t getByteFromOpcode(uint16_t opcode) { return opcode & 0x00FF; }
inline uint16_t getWordFromOpcode(uint16_t opcode) { return opcode & 0x0FFF; }

#define NEXT_INSTRUCTION() m_currentState.PC += 2;

namespace Chip8 {

const char* StatusNames[4] = {
	"Reset", "Initialized", "Running", "Stopped"
};

const char* KeyNames[16] = {
	"KP1", "KP2", "KP3", "KPC", "KP4", "KP5", "KP6", "KPD", "KP7", "KP8", "KP9", "KPE", "KPA", "KP0", "KPB", "KPF"
};

const uint8_t Chip8KeypadToIntTable[] = {
	13, 0, 1, 2, 4, 5, 6, 8, 9, 10, 12, 14, 3, 7, 11, 15
};

const uint8_t IntToChip8KeypadTable[] = {
	1, 2, 3, 12, 4, 5, 6, 13, 7, 8, 9, 14, 10, 0, 11, 15
};

VirtualMachine::VirtualMachine() : m_timingTable(CycleTiming::Default), m_appFrontend(0), m_debugger(0)
{
	reset();
}

VirtualMachine::VirtualMachine(IApplicationFrontend* af) : m_timingTable(CycleTiming::Default), m_appFrontend(af), m_debugger(0)
{
	reset();
}

void VirtualMachine::load(Program* program)
{
	m_program = program;

	restart();
	/*switch (program.type)
	{
		case TYPE_CHIP8:
			
			
			break;
		case TYPE_SCHIP:
			break;
		case TYPE_MEGACHIP:
			break;
	}*/
}

void VirtualMachine::update(uint64_t dt)
{
	if (m_currentState.status != Status::Stopped)
	{
		//we will ignore the cycles that are used for the fetch-decode part

		//count down timers at 60 Hz (once every 16.67 ms)
		//Chip8 at ~250 Hz and SChip at 1000 Hz
		//cycles mean machine cycles, and then we count down by instruction cycles

		//calculate how many cycles to go through based on the delta time passed in
		float dt_seconds = (dt / 1000000.f);
		float cyclesToProcess = dt_seconds * (m_currentState.frequency);
		//m_currentState.videoBuffer[0] = 100;

		float timePerCycle = dt_seconds / cyclesToProcess;
		if (m_appFrontend)
		{
			m_appFrontend->clearLog();
			/*m_appInterface->logMessage("dt: %d", dt);
			m_appInterface->logMessage("dt: %f", dt_seconds);
			m_appInterface->logMessage("cycles/sec: %d", m_currentState.frequency);
			m_appInterface->logMessage("cyclesToProcess: %f", cyclesToProcess);
			m_appInterface->logMessage("cycleAccumulator: %f", m_currentState.cycleAccumulator);
			m_appInterface->logMessage("timePerCycle: %f", timePerCycle);
			m_appInterface->logMessage("timerAccumulator: %f", m_currentState.timerAccumulator);
			m_appInterface->logMessage("Timers: ");
			m_appInterface->logMessage("\tdelay: %d", m_currentState.timerDelay);
			m_appInterface->logMessage("\tsound: %d", m_currentState.timerSound);*/
			logExecutionInfo();
		}
	
		m_currentState.cycleAccumulator += cyclesToProcess;
	
		while (m_currentState.cycleAccumulator > 0)
		{
			updateTimers(timePerCycle);

			fetchAndDecode();
		
			//when we don't have enough cycles to process during this update(), we will still complete one full opcode
			//and then remember how much we went over, to use that time as waiting later
			//cyclesToProcess -= (opcode cycle usage);
			//hardcoded that every opcode takes 1 cycle currently
			m_currentState.cycleAccumulator -= 1;//(opcode cycle usage);
		}
	}
}



void VirtualMachine::step(uint64_t dt)
{
	//step() should be possible from a debugger, so we don't check against Stopped status here,
	if (m_appFrontend)
		m_appFrontend->clearLog();

	logExecutionInfo();

	//when we do single-stepping, just add all the delta time, and decrement the timers accordingly
	updateTimers(dt);

	fetchAndDecode();
		
	//when we don't have enough cycles to process during this update(), we will still complete one full opcode
	//and then remember how much we went over, to use that time as waiting later
	//cyclesToProcess -= (opcode cycle usage);
	//hardcoded that every opcode takes 1 cycle currently
	m_currentState.cycleAccumulator -= 1;//(opcode cycle usage);

}

Program* VirtualMachine::getProgram()
{
	return m_program;
}

void VirtualMachine::restart()
{
	reset();

	//for now, we will always go for 0x200 as the start of execution
	//there are certain programs that start at 0x600 and 0x260, but we will ignore them for now
	m_currentState.PC = 0x200;

	memcpy(&m_currentState.memory[0x200], &m_program->bytecode[0], m_program->bytecode.size());

	m_currentState.status = Status::Initialized;
}

const VirtualMachine::HardwareOutput& VirtualMachine::getHardwareOutput()
{
	m_hardwareOutput.videoBuffer = m_currentState.videoBuffer;

	//set the video buffer size depending on what mode we are in
	if (!m_currentState.superChipMode && !m_currentState.megaChipMode)
	{
		m_hardwareOutput.width = 128; //64;
		m_hardwareOutput.height = 64; //32;
	} else if (m_currentState.superChipMode && !m_currentState.megaChipMode)
	{
		m_hardwareOutput.width = 128;
		m_hardwareOutput.height = 64;
	} else
	{
		m_hardwareOutput.width = 256;
		m_hardwareOutput.height = 192;
	}

	m_hardwareOutput.makeNoise = m_currentState.timerSound > 0 ? true : false;

	return m_hardwareOutput;
}

const std::list<VirtualMachine::OpcodeEntry>& VirtualMachine::getExecutionHistory() const
{
	return m_vmhistory;
}

VirtualMachine::State& VirtualMachine::saveState()
{
	return m_currentState;
}

void VirtualMachine::loadState(const State& state)
{
	//reset();
	memset(m_keypadActivity, 0, 0x10);
	m_currentState = state;
}

void VirtualMachine::attachDebugger(IDebugger* debugger)
{
	m_debugger = debugger;
	m_debugger->onAttach(this);
}

void VirtualMachine::reset()
{
	/*
	The value we set the PC to at start depends on what type of program we are running, so it will be set again in load()
	0x600 (sometimes for the ETI 660 computer) may be the exception if we can find any program using it
	Update: the Chip8 program pack from chip8.com includes both ETI660 programs and chip8 hires programs
	chip8 hires programs start at 0x260
	*/
	m_currentState.PC = 0;

	memset(m_currentState.memory, 0, 0xFFF);

	memcpy(m_currentState.memory, gChip8Fonts, 80);
	memcpy(&m_currentState.memory[0x51], gSChipFonts, 160);

	memset(m_currentState.registerV, 0, 0x10);

	m_currentState.registerI = 0;
	m_currentState.timerDelay = 0;
	m_currentState.timerSound = 0;

	memset(m_currentState.stack, 0, 0xF);
	m_currentState.stackPointer = 0;

	m_currentState.superChipMode = false;
	m_currentState.megaChipMode = false;
	m_currentState.correctChip8Shift = false;

	memset(m_currentState.HP48UserFlag, 0, 8);
	m_currentState.HP48UserFlag[1] = 0xAA;
	m_currentState.HP48UserFlag[2] = 0xAA;
	memset(m_currentState.videoBuffer, 0, 256*192);

	m_currentState.timerAccumulator = 0;
	m_currentState.cycleAccumulator = 0;
	m_currentState.frequency = CycleTiming::DefaultVMFrequency;
	m_currentState.status = Status::Reset;

	memset(m_keypadActivity, 0, sizeof(bool)*0x10);
	//TODO: use different RNG
	srand(time(0));
}

void VirtualMachine::logExecutionInfo()
{
	if (m_appFrontend)
	{
		if (m_currentState.superChipMode)
			m_appFrontend->logMessage("Mode: SuperChip8");
		else
			m_appFrontend->logMessage("Mode: Chip8");
		m_appFrontend->logMessage("Status: %s", StatusNames[static_cast<int>(m_currentState.status)]);
		/*m_appInterface->logMessage("PC: %#4x", m_currentState.PC);
		m_appInterface->logMessage("Register data -----");
		m_appInterface->logMessage("\tI: %x", m_currentState.registerI);
		for (int i = 0; i < 16; ++i)
			m_appInterface->logMessage("\tV%d: %x", i, m_currentState.registerV[i]);*/
	}
}

void VirtualMachine::fetchAndDecode()
{
	uint16_t oldPC = m_currentState.PC;
	uint16_t opcode = ((m_currentState.memory[m_currentState.PC] << 8) & 0xFFFF) + (m_currentState.memory[m_currentState.PC + 1] & 0xFF);

	if (m_debugger)
	{
		bool ret = m_debugger->onStep(opcode);
		if (!ret)
		{
			m_currentState.status = Status::Stopped;
			return;
		}
	}

	interpretOpcode(opcode);

	if (m_currentState.PC != oldPC)
	{
		//if we had a change of PC, we can assume something happened, if not, we have probably entered an infinite loop
		/*if (m_appInterface)
			m_appInterface->logMessage("opcode: %#4x", opcode);*/
		if (m_currentState.status != Status::Stopped && m_currentState.status != Status::Running)
			m_currentState.status = Status::Running;

		m_vmhistory.push_back(OpcodeEntry(oldPC, opcode));
		if (m_vmhistory.size() > 100)
			m_vmhistory.pop_front();
	} else
	{
		//we jumped back to previous location, infinite loop going on
		//throw std::exception("loopu");
	}
}

void VirtualMachine::updateTimers(float dt)
{
	m_currentState.timerAccumulator += dt;
	while (m_currentState.timerAccumulator >= 0.016667f)
	{
		m_currentState.timerAccumulator -= 0.016667f;
		if (m_currentState.timerDelay > 0)
			m_currentState.timerDelay--;
		if (m_currentState.timerSound > 0)
		{
			m_currentState.timerSound--;
			m_hardwareOutput.makeNoise = true;
		} else
		{
			m_hardwareOutput.makeNoise = false;
		}
	}
}

void VirtualMachine::interpretOpcode(uint16_t opcode)
{

	switch (opcode & 0xF000)
	{
		case 0x0000:
			handleSuperChip(opcode);
			break;
		case 0x1000:
			m_currentState.PC = getWordFromOpcode(opcode);
			break;
		case 0x2000:
			assert(m_currentState.stackPointer < 0xF);
			m_currentState.stack[m_currentState.stackPointer++] = m_currentState.PC;
			m_currentState.PC = getWordFromOpcode(opcode);
			break;
		case 0x3000:
			if (m_currentState.registerV[ getXFromOpcode(opcode) ] == getByteFromOpcode(opcode))
				NEXT_INSTRUCTION();

			NEXT_INSTRUCTION();
			break;
		case 0x4000:
			if (m_currentState.registerV[ getXFromOpcode(opcode) ] != getByteFromOpcode(opcode))
				NEXT_INSTRUCTION();

			NEXT_INSTRUCTION();
			break;
		case 0x5000:
			if (m_currentState.registerV[ getXFromOpcode(opcode) ] == m_currentState.registerV[ getYFromOpcode(opcode) ])
				NEXT_INSTRUCTION();

			NEXT_INSTRUCTION();
			break;
		case 0x6000:
			m_currentState.registerV[ getXFromOpcode(opcode) ] = getByteFromOpcode(opcode);
			NEXT_INSTRUCTION();
			break;
		case 0x7000:
			m_currentState.registerV[ getXFromOpcode(opcode) ] += getByteFromOpcode(opcode);
			NEXT_INSTRUCTION();
			break;
		case 0x8000:
			handleArithmeticAndLogic(opcode);
			break;
		case 0x9000:
			if (m_currentState.registerV[ getXFromOpcode(opcode) ] != m_currentState.registerV[ getYFromOpcode(opcode) ])
				NEXT_INSTRUCTION();

			NEXT_INSTRUCTION();
			break;
		case 0xA000:
			m_currentState.registerI = getWordFromOpcode(opcode) & 0xFFF;
			NEXT_INSTRUCTION();
			break;
		case 0xB000:
			throw std::exception("opcode not supported yet");
			break;
		case 0xC000:
			//random number hurrah
			//we need to be careful how we generate numbers here
			//TODO: do we need to provide option to switch between & and %??
			//UPDATE: it would seem & is the good option for everything, just like documentation states
			//most likely, emulators that get this wrong forget to do the rand() % 255 first
			m_currentState.registerV[ getXFromOpcode(opcode) ] = (rand() % 255) & getWordFromOpcode(opcode);
			NEXT_INSTRUCTION();
			break;
		case 0xD000:
			handleDraw(opcode);
			NEXT_INSTRUCTION();
			break;
		case 0xE000:
			handleKeypress(opcode);
			break;
		case 0xF000:
			handleFOpcodes(opcode);
			break;
	}

}

void VirtualMachine::handleSuperChip(uint16_t opcode)
{
	switch (opcode & 0x00FF)
	{
		case 0xFB:
			handleScrollRight();
			NEXT_INSTRUCTION();
			break;
		case 0xFC:
			handleScrollLeft();
			NEXT_INSTRUCTION();
			break;
		case 0xFD:
			m_currentState.status = Status::Stopped;
			break;
		case 0xFE:
			m_currentState.superChipMode = false;
			NEXT_INSTRUCTION();
			break;
		case 0xFF:
			m_currentState.superChipMode = true;
			NEXT_INSTRUCTION();
			break;
		case 0xE0:
			memset(m_currentState.videoBuffer, 0, VideoBufferSizeMegaChip);
			NEXT_INSTRUCTION();
			break;
		case 0xEE:
			assert(m_currentState.stackPointer > 0);
			m_currentState.PC = m_currentState.stack[--m_currentState.stackPointer];
			//Note: after returning from a subroutine call, we of course need to move to the next instruction after the calling point
			//otherwise we get into an infinite loop
			NEXT_INSTRUCTION();
			break;
		default:
			if ((opcode & 0x00F0) == 0xC0)
			{
				//0x00CN
				//scroll video buffer N lines down
				handleScrollDown(opcode & 0x000F);
				NEXT_INSTRUCTION();
			} else
			{
				//if we get here, we have eliminated all other possibilites, so we can assume it's an 0NNN opcode, which
				//is used by hybrid Chip8 programs that call 1802 programs
				throw std::exception("do not support calling arbitrary 1802 program code");
			}
			break;
	}
}

void VirtualMachine::handleArithmeticAndLogic(uint16_t opcode)
{
	

	switch (opcode & 0x000F)
	{
		case 0x0:
			m_currentState.registerV[ getXFromOpcode(opcode) ] = m_currentState.registerV[ getYFromOpcode(opcode) ];
			NEXT_INSTRUCTION();
			break;
		case 0x1:
			m_currentState.registerV[ getXFromOpcode(opcode) ] |= m_currentState.registerV[ getYFromOpcode(opcode) ];
			NEXT_INSTRUCTION();
			break;
		case 0x2:
			m_currentState.registerV[ getXFromOpcode(opcode) ] &= m_currentState.registerV[ getYFromOpcode(opcode) ];
			NEXT_INSTRUCTION();
			break;
		case 0x3:
			m_currentState.registerV[ getXFromOpcode(opcode) ] ^= m_currentState.registerV[ getYFromOpcode(opcode) ];
			NEXT_INSTRUCTION();
			break;
		case 0x4:
			{
				//TODO: see if we can calculate the carry flag without using a 16bit temporary?
				uint16_t tempResult = m_currentState.registerV[ getXFromOpcode(opcode) ] + m_currentState.registerV[ getYFromOpcode(opcode) ];
				m_currentState.registerV[ getXFromOpcode(opcode) ] += m_currentState.registerV[ getYFromOpcode(opcode) ];
				m_currentState.registerV[0xF] = tempResult > 255 ? 1 : 0;
				NEXT_INSTRUCTION();
				break;
			}
		case 0x5:
		{
			uint8_t borrow = (m_currentState.registerV[ getXFromOpcode(opcode) ] < m_currentState.registerV[ getYFromOpcode(opcode) ] ? 0 : 1);
			m_currentState.registerV[ getXFromOpcode(opcode) ] -= m_currentState.registerV[ getYFromOpcode(opcode) ];
			m_currentState.registerV[0xF] = borrow;
			NEXT_INSTRUCTION();
			break;
		}
		case 0x6: //Vx shiftright 1 step VF carry
			/*
			Interesting information here: http://laurencescotford.co.uk/?p=266
			This actually shifts Vy, and not Vx. But a lot of games may have been made under the wrong assumption...
			*/

			if (m_currentState.correctChip8Shift)
			{
				uint8_t carry = m_currentState.registerV[ getYFromOpcode(opcode) ] & 0x1;
				//shift Vy right 1 step and set the result to Vx
				m_currentState.registerV[ getXFromOpcode(opcode) ] = m_currentState.registerV[ getYFromOpcode(opcode) ] >> 1;
				m_currentState.registerV[0xF] = carry;
			} else
			{
				//the non-standard behaviour is instead to shift Vx and save the result back to Vx
				//TODO/NOTE: even though we modify the same register that we store in, we may still want to move the carry calc to after
				m_currentState.registerV[0xF] = m_currentState.registerV[ getXFromOpcode(opcode) ] & 0x1;
				//shift Vy right 1 step and set the result to Vx
				m_currentState.registerV[ getXFromOpcode(opcode) ] = m_currentState.registerV[ getXFromOpcode(opcode) ] >> 1;
			}
			NEXT_INSTRUCTION();
			break;
		case 0x7:
		{
			uint8_t borrow = (m_currentState.registerV[ getXFromOpcode(opcode) ] <= m_currentState.registerV[ getYFromOpcode(opcode) ] ? 1 : 0);
			m_currentState.registerV[ getXFromOpcode(opcode) ] = m_currentState.registerV[ getYFromOpcode(opcode) ] - m_currentState.registerV[ getXFromOpcode(opcode) ];
			m_currentState.registerV[0xF] = borrow;
			NEXT_INSTRUCTION();
			break;
		}
		case 0xE: //Vx shiftleft 1 step, VF = carry
			//See 0x6 for more info on the misconception about the shift instructions
			if (m_currentState.correctChip8Shift)
			{
				uint8_t carry = (m_currentState.registerV[ getYFromOpcode(opcode) ] >> 7) & 0x1;
				m_currentState.registerV[ getXFromOpcode(opcode) ] = m_currentState.registerV[ getYFromOpcode(opcode) ] << 1;
				m_currentState.registerV[0xF] = carry;
			} else
			{
				m_currentState.registerV[0xF] = (m_currentState.registerV[ getXFromOpcode(opcode) ] >> 7) & 0x1;
				m_currentState.registerV[ getXFromOpcode(opcode) ] = m_currentState.registerV[ getXFromOpcode(opcode) ] << 1;
			}
			NEXT_INSTRUCTION();
			break;
		default:
			throw std::exception("found unhandled ALU instruction");
			break;
	}
}

void VirtualMachine::handleDraw(uint16_t opcode)
{
	/*
	David Winter's documentation is the only one I've seen that mentions any details about the SuperChip vertical scrolling's half-pixel shift.
	What it really means is that a request can be made to scroll 1 line down, which would then be 1/2 line in Chip8 graphics resolution.
	To easily handle this here, we will instead render all Chip8 graphics in SuperChip resolution, only enlargened so it looks the same.
	*/

	uint8_t x1 = m_currentState.registerV[ getXFromOpcode(opcode) ];
	uint8_t y1 = m_currentState.registerV[ getYFromOpcode(opcode) ];

	if (m_currentState.superChipMode)
	{
		assert(x1 < 128);
		assert(y1 < 64);
	} /*else
	{
		assert(x1 < 64);
		assert(y1 < 32);
	}*/

	uint8_t spriteHeight = opcode & 0x000F;

	if (spriteHeight == 0)
		spriteHeight = 16;

	//collision flag is supposed to be reset before the drawing occurs
	m_currentState.registerV[0xF] = 0;

	if (m_currentState.superChipMode)
	{
		//in superchip mode, we either want to draw a 16x16 sprite, or an NxM (N,M N <= 8, M <= 15), in high resolution
		handleDrawSuperChipSprite(x1, y1, spriteHeight);
	} else
	{
		//8x16 lowres, or (N,M N <= 8, M <= 15) lowres (but both scaled into highres to support scrolling)
		handleDrawLowresSprite(x1, y1, spriteHeight);
	}
}

void VirtualMachine::handleDrawSuperChipSprite(uint8_t x, uint8_t y, uint8_t height)
{
	uint8_t xpos, ypos;
	if (height == 16)
	{
		for (uint8_t row = 0, rowByte = 0; row < height; ++row, rowByte += 2)
		{
			ypos = (y + (row));// % 32;

			for (uint8_t columnBit = 0; columnBit < 8; ++columnBit)
			{
				xpos = (x + columnBit);// % 64;

				if (m_currentState.memory[m_currentState.registerI + rowByte] & (0x80 >> columnBit))
				{
					if (m_currentState.videoBuffer[xpos + ypos * 128] == 1)
						m_currentState.registerV[0xF] = 1;

					m_currentState.videoBuffer[xpos + ypos * 128] ^= 1;
				}
			}
			//draw the second byte
			
			for (uint8_t columnBit = 0; columnBit < 8; ++columnBit)
			{
				xpos = (x + 8 + columnBit);// % 64;

				if (m_currentState.memory[m_currentState.registerI + rowByte + 1] & (0x80 >> columnBit))
				{
					if (m_currentState.videoBuffer[xpos + ypos * 128] == 1)
						m_currentState.registerV[0xF] = 1;

					m_currentState.videoBuffer[xpos + ypos * 128] ^= 1;
				}
			}

		}
	} else
	{
		for (uint8_t row = 0; row < height; ++row)
		{
			ypos = (y + row);// % 32;

			for (uint8_t columnBit = 0; columnBit < 8; ++columnBit)
			{
				xpos = (x + columnBit);// % 64;

				if (m_currentState.memory[m_currentState.registerI + row] & (0x80 >> columnBit))
				{
					if (m_currentState.videoBuffer[xpos + ypos * 128] == 1)
						m_currentState.registerV[0xF] = 1;

					m_currentState.videoBuffer[xpos + ypos * 128] ^= 1;
				}
			}
		}
	}
}

void VirtualMachine::handleDrawLowresSprite(uint8_t x, uint8_t y, uint8_t height)
{
	const uint8_t screenWidth = 128;

	uint8_t xpos, ypos;

	if (height == 16)
	{
		for (uint8_t row = 0; row < height; ++row)
		{
			ypos = (y + row)*2;// % 32;

			for (uint8_t columnBit = 0; columnBit < 8; ++columnBit)
			{
				xpos = (x + columnBit)*2;// % 64;

				if (m_currentState.memory[m_currentState.registerI + row] & (0x80 >> columnBit))
				{
					if (m_currentState.videoBuffer[xpos + ypos * screenWidth] == 1)
						m_currentState.registerV[0xF] = 1;

					m_currentState.videoBuffer[xpos + ypos * screenWidth] ^= 1;
					m_currentState.videoBuffer[xpos + 1 + ypos * screenWidth] ^= 1;
					m_currentState.videoBuffer[xpos + 1 + (ypos + 1) * screenWidth] ^= 1;
					m_currentState.videoBuffer[xpos + (ypos + 1) * screenWidth] ^= 1;
				}
			}
		}
	} else
	{
		for (uint8_t row = 0; row < height; ++row)
		{
			ypos = (y + row) * 2;// % 32;

			for (uint8_t columnBit = 0; columnBit < 8; ++columnBit)
			{
				xpos = (x + columnBit) * 2;// % 64;

				if (m_currentState.memory[m_currentState.registerI + row] & (0x80 >> columnBit))
				{
					if (m_currentState.videoBuffer[xpos + ypos * screenWidth] == 1)
						m_currentState.registerV[0xF] = 1;

					m_currentState.videoBuffer[xpos + ypos * screenWidth] ^= 1;
					m_currentState.videoBuffer[xpos + 1 + ypos * screenWidth] ^= 1;
					m_currentState.videoBuffer[xpos + 1 + (ypos + 1) * screenWidth] ^= 1;
					m_currentState.videoBuffer[xpos + (ypos + 1) * screenWidth] ^= 1;
				}
			}
		}
	}
}

void VirtualMachine::handleScrollDown(uint8_t lines)
{
	const size_t lineSize = 128;//m_currentState.superChipMode ? 128 : 64;

	//for (uint8_t y = m_currentState.superChipMode ? 63 : 31; y >= lines; --y)
	for (uint8_t y = 63; y >= lines; --y)
	{
		memcpy(&m_currentState.videoBuffer[y * lineSize], &m_currentState.videoBuffer[ (y - lines) * lineSize], lineSize);
	}

	for (uint8_t i = 0; i < lines; ++i)
	{
		memset(&m_currentState.videoBuffer[ i * lineSize], 0, lineSize);
	}
}

void VirtualMachine::handleScrollRight()
{
	const uint8_t scrollAmount = 4; //m_currentState.superChipMode ? 4 : 2;
	const size_t lineSize = 128; //m_currentState.superChipMode ? 128 : 64;
	const uint8_t height = 64; //m_currentState.superChipMode ? 64 : 32;

	for (uint8_t y = 0; y < height; ++y)
	{
		memmove(&m_currentState.videoBuffer[y * lineSize + scrollAmount], &m_currentState.videoBuffer[y * lineSize], lineSize - scrollAmount);
	}

	for (uint8_t y = 0; y < height; ++y)
	{
		memset(&m_currentState.videoBuffer[y * lineSize], 0, scrollAmount);
	}
}

void VirtualMachine::handleScrollLeft()
{
	const uint8_t scrollAmount = 4; //m_currentState.superChipMode ? 4 : 2;
	const size_t lineSize = 128; //m_currentState.superChipMode ? 128 : 64;
	const uint8_t height = 64; //m_currentState.superChipMode ? 64 : 32;

	for (uint8_t y = 0; y < height; ++y)
	{
		memmove(&m_currentState.videoBuffer[y * lineSize], &m_currentState.videoBuffer[y * lineSize + scrollAmount], lineSize - scrollAmount);
	}

	for (uint8_t y = 0; y < height; ++y)
	{
		memset(&m_currentState.videoBuffer[lineSize - scrollAmount + y * lineSize], 0, scrollAmount);
	}

}

void VirtualMachine::handleKeypress(uint16_t opcode)
{
	
	switch (opcode & 0x00FF)
	{
		case 0x9E:
			assert(m_currentState.registerV[ getXFromOpcode(opcode) ] < 0x10);
			assert(m_appFrontend != nullptr);

			m_keypadActivity[ Chip8KeypadToIntTable[m_currentState.registerV[ getXFromOpcode(opcode) ]] ] = true;

			if (m_appFrontend->getKeyDown(static_cast<Key>( Chip8KeypadToIntTable[m_currentState.registerV[ getXFromOpcode(opcode) ]] )))
				NEXT_INSTRUCTION();

			NEXT_INSTRUCTION();
			break;
		case 0xA1:
			assert(m_currentState.registerV[ getXFromOpcode(opcode) ] < 0x10);
			assert(m_appFrontend != nullptr);

			m_keypadActivity[ Chip8KeypadToIntTable[m_currentState.registerV[ getXFromOpcode(opcode) ]] ] = true;

			if (!m_appFrontend->getKeyDown(static_cast<Key>( Chip8KeypadToIntTable[m_currentState.registerV[ getXFromOpcode(opcode) ]] )))
				NEXT_INSTRUCTION();

			NEXT_INSTRUCTION();
			break;
	}
}

void VirtualMachine::handleFOpcodes(uint16_t opcode)
{
	
	switch (opcode & 0x00FF)
	{
		case 0x07: //regVx = delaytimer
			m_currentState.registerV[ getXFromOpcode(opcode) ] = m_currentState.timerDelay;
			NEXT_INSTRUCTION();
			break;
		case 0x0A:
			assert(m_appFrontend != nullptr);
			//wait for keypress before advancing PC
			//any keypress is fine, but we have to wait for this
			for (uint8_t i = 0; i < 0x10; ++i)
			{
				if (m_appFrontend->getKeyDown(static_cast<Key>(i)))
				{
					m_currentState.registerV[ getXFromOpcode(opcode) ] = IntToChip8KeypadTable[i];
					NEXT_INSTRUCTION();
					break;
				}
			}
			break;
		case 0x15: //delay timer = regVx
			m_currentState.timerDelay = m_currentState.registerV[ getXFromOpcode(opcode) ];
			NEXT_INSTRUCTION();
			break;
		case 0x18: //soundtimer = regVx
			m_currentState.timerSound = m_currentState.registerV[ getXFromOpcode(opcode) ];
			NEXT_INSTRUCTION();
			break;
		case 0x1E:
			{
				bool overflow = false;
				//This is an undocumented feature of this instruction. If the result overflows, VF should be set to 1, otherwise 0.
				//The only known game that relies on this is Spacefight 2019. See documentation for SC Test for more information.
				if ((m_currentState.registerI + m_currentState.registerV[ getXFromOpcode(opcode) ]) > 4095)
					overflow = true;
				m_currentState.registerI += m_currentState.registerV[ getXFromOpcode(opcode) ];
				NEXT_INSTRUCTION();
			
				m_currentState.registerV[0xF] = overflow;
				break;
			}
		case 0x29:
			m_currentState.registerI = m_currentState.registerV[ getXFromOpcode(opcode) ] * 5;
			NEXT_INSTRUCTION();
			break;
		case 0x30:
			m_currentState.registerI = 81 + m_currentState.registerV[ getXFromOpcode(opcode) ] * 10;
			NEXT_INSTRUCTION();
			break;
		case 0x33:
			{
				//store BCD of register V[x] on memory
				assert(m_currentState.registerI + 2 < 0xFFF);
				uint8_t x = m_currentState.registerV[ getXFromOpcode(opcode) ];
				m_currentState.memory[m_currentState.registerI] = x / 100;
				m_currentState.memory[m_currentState.registerI + 1] = (x / 10) % 10;
				m_currentState.memory[m_currentState.registerI + 2] = x % 10;
				NEXT_INSTRUCTION();
				break;
			}
		case 0x55:
			{
				uint8_t registersToCopy = getXFromOpcode(opcode) + 1;
				assert((m_currentState.registerI + registersToCopy) < 0xFFF);

				memcpy(&m_currentState.memory[m_currentState.registerI], m_currentState.registerV, registersToCopy);
				NEXT_INSTRUCTION();
				break;
			}
		case 0x65:
			{
				//
				uint8_t registersToCopy = getXFromOpcode(opcode);
				assert((m_currentState.registerI + registersToCopy) < 0xFFF);
				//if( (m_currentState.registerI + registersToCopy) > 0xFFF)
				//	registersToCopy -= (0xFFF - m_currentState.registerI);

				memcpy(m_currentState.registerV, &m_currentState.memory[m_currentState.registerI], registersToCopy + 1);
				//TODO: verify if registerI should be increased or not
				//NOTE: the SuperChip game Ant breaks if we increase I here, maybe others too
				//m_currentState.registerI += registersToCopy;
				NEXT_INSTRUCTION();
				break;
			}
		case 0x75:
			assert(getXFromOpcode(opcode) <= 7);
			memcpy(m_currentState.HP48UserFlag, m_currentState.registerV, getXFromOpcode(opcode) + 1);
			NEXT_INSTRUCTION();
			break;
		case 0x85:
			assert(getXFromOpcode(opcode) <= 7);
			memcpy(m_currentState.registerV, m_currentState.HP48UserFlag, getXFromOpcode(opcode) + 1);
			NEXT_INSTRUCTION();
			break;
		default:
			throw std::exception("F0NN not supported yet");
			break;
	}
}

} /* end namespace */