/*
Some random Chip8 tidbits for later:

"
 Originally Posted by hap 
It's due to the emulated CPU speed.

 Hidden runs nice at 0.1KHz
 Tetris runs nice at 0.3KHz
 Brix runs nice at 0.5KHz
 and S-chip games usually 'need' over 1KHz.
"

On the random generator:
For Car, Race, Tetris, take a look at your random opcode. Make sure the random number max value is at least 255, and min is 0, and it should be ANDed with the low byte of the opcode.
"""""""""""""""""""""""""""
Isn't it the biggest difference between them?  Tetris doesn't work properly if the method is %, while rocket isn't generating no starfield if it is &. 
Some other games depend on this random stuff too. Like those car games. The other solution is to fix the code in the roms, so all of them use the right method for this. 
But this would get problematically, because there are a lot of sources to get the same game roms from, so replacing them with fixed ones would be nearly impossible.
"""""""""""""""""""""""""""
Memory:

CHIP-8's memory addresses range from 200h to FFFh, making for 3,584 bytes. The reason for the memory starting at 200h is that on the Cosmac 
VIP and Telmac 1800, the first 512 bytes are reserved for the interpreter. On those machines, the uppermost 256 bytes (F00h-FFFh on a 4K machine) 
were reserved for display refresh, and the 96 bytes below that (EA0h-EFFh) were reserved for the call stack, internal use, and the variables.

Graphics idea:
Monitor how frequently a certain position in the buffer is modified, and adjust a grayscale value depending on this.


For MegaChip drawing:

http://chip8.wikia.com/wiki/Instruction_Draw
http://chip8.wikia.com/wiki/Instruction_Load_Palette


Original Chip8 has 35 instructions.
SuperChip8 adds 9, for a total of 44 instructions
MegaChip further adds 11, totaling out at 55 instructions (including higher res color graphics)


A thread that could be useful:
http://ngemu.com/threads/chip8-thread.114578/page-13


*/

#pragma once

#include <vector>
#include <list>
#include <cstdint>
#include <sstream>



namespace Chip8 {

enum Chip8ProgramMode {
	MODE_CHIP8,
	MODE_SCHIP,
	MODE_MEGACHIP
};


const int VideoBufferSizeChip8 = 64*32; //ratio 2
const int VideoBufferSizeSuperChip = 128*64; //ratio 2
const int VideoBufferSizeMegaChip = 256*192; //ratio 1.333


extern const char* StatusNames[4];
extern const char* KeyNames[16];

enum class Key : uint8_t {
	KP0, KP1, KP2, KP3, KP4, KP5, KP6, KP7, KP8, KP9, KPA, KPB, KPC, KPD, KPE, KPF
};

struct Program
{
	std::vector<uint8_t> bytecode;
	uint32_t desiredFrequency;
};

class IDebugger
{
	public:
		virtual void onAttach(class VirtualMachine*) = 0;
		virtual bool onStep(uint16_t opcode) = 0;

	private:
};

class VirtualMachine
{
	public:
		class IApplicationFrontend
		{
			public:
				//TODO: add keyboard input stuff
				virtual bool getKeyDown(Key key) = 0;

				//virtual void logMessage(const std::string& msg) = 0;
				virtual void logMessage(const char* format, ...) = 0;
				
				virtual void clearLog() = 0;
		};

		VirtualMachine();
		VirtualMachine(IApplicationFrontend* ai);

		void load(Program* program);

		/*

		dt = delta time in microseconds
		*/
		void update(uint64_t dt);

		/*
		Use step() as a way to advance the VM one instruction at a time.
		It still requires a delta time parameter so we can properly update the timers.
		*/
		void step(uint64_t dt);

		void restart();

		Program* getProgram();

		struct HardwareOutput
		{
			const uint8_t* videoBuffer;
			uint16_t width;
			uint16_t height;
			bool makeNoise; //the chip8 beep
		};	

		const HardwareOutput& getHardwareOutput();

		enum class Status : int {
			Reset, Initialized, Running, Stopped
		};

		Status getExecuteStatus() { return m_currentState.status; }

		struct State
		{
			
			//the chip8 VM hardware emulation stuff
			uint16_t		PC;
			uint8_t			memory[0xFFF];
			uint8_t			registerV[0x10];
			uint16_t		registerI;

			/*
			David Winter's documentation states that the two timers are 8-bit, but all other doc's seem to not say anything on the size.
			They count down at 60Hz
			*/
			uint32_t		timerDelay;
			uint32_t		timerSound;

			uint16_t		stack[0xF]; //hardcoded for 16 stack levels
			uint8_t			stackPointer;

			bool			superChipMode;
			bool			megaChipMode;
			bool			correctChip8Shift; //whether we should use the correct shift operations (shift Vy) or use the wildy believed shift Vx which lots of programs rely on

			/*
			Extension added by the HP48 graphing calculator, and thus only used in SCHIP mode (supposedly)
			Some documentation says the opcodes that modify this only allow modification of up to 8 of these. There are 16 V registers, but perhaps
			you are only allowed to store at most 8 at a time using this.
		
			Some info on HP48 and it's user flags:
				http://documents.epfl.ch/users/f/fr/froulet/www/HP/RPL/usr_cond.html
		
			TODO: change to 16 if this is ever needed
			*/
			uint8_t			HP48UserFlag[8]; 

			uint8_t			videoBuffer[VideoBufferSizeMegaChip]; //wasteful way of declaring the array, but avoids heap allocation
			/*uint16_t		videoWidth;
			uint16_t		videoHeight;*/

			//VM internal working stuff
			float timerAccumulator;
			float cycleAccumulator;
			uint32_t frequency;
			Status status;
		};

		State& saveState();

		const State& getCurrentState() const { return m_currentState; }
		/*
		Because the program code from the file is not stored internally, but only referenced by a pointer, it is not guaranteed that you will be able to perform a restart() on the VM after loading
		a saved state.
		*/
		void loadState(const State& state);

		struct OpcodeEntry
		{
			OpcodeEntry(uint16_t inPC, uint16_t inOpcode) : pc(inPC), opcode(inOpcode) {}
		
			uint16_t pc;
			uint16_t opcode;
		};

		const std::list<OpcodeEntry>& getExecutionHistory() const;
		const bool*	getKeypadActivity() const { return m_keypadActivity; }

		void attachDebugger(IDebugger* debugger);

	private:
		void reset();
		
		void logExecutionInfo();

		void fetchAndDecode();
		void updateTimers(float dt);
		void interpretOpcode(uint16_t opcode);
		void handleSuperChip(uint16_t opcode);
		void handleArithmeticAndLogic(uint16_t opcode);
		void handleDraw(uint16_t opcode);
		void handleDrawSuperChipSprite(uint8_t x, uint8_t y, uint8_t height);
		void handleDrawLowresSprite(uint8_t x, uint8_t y, uint8_t height);

		void handleScrollDown(uint8_t lines);
		void handleScrollRight();
		void handleScrollLeft();
		void handleKeypress(uint16_t opcode);
		void handleFOpcodes(uint16_t opcode);
		

		//void updateTimers(uint64_t dt);
		IApplicationFrontend*	m_appFrontend;
		IDebugger*				m_debugger;

		Program*		m_program;
		const uint32_t*	m_timingTable;

		HardwareOutput	m_hardwareOutput;
		State			m_currentState;

		std::list<OpcodeEntry>	m_vmhistory;
		bool					m_keypadActivity[0x10];
};



} /* end namespace Chip8 */