#include "CustomSFMLGUI.h"

namespace UI {

namespace Default {

	void Initialize(const std::string& fontFile)
	{
		TextFont = new sf::Font();
		TextFont->loadFromFile(fontFile);
	}

	/*
	We use a pointer rather than a stack allocated instance because of a problem with the design of sf::Font, in it's deallocation.
	It will cause the application to cras on exit unless we use a pointer.
	One issue with using the pointer here is that we never explicitly deallocate it now, it will get cleaned up by
	application exit though, so it's not THAT bad...
	*/
	sf::Font* TextFont;
};


sf::Vector2f minVector(const sf::Vector2f& a, const sf::Vector2f& b)
{
	if (a.x <= b.x && a.y <= b.y)
		return a;
	else
		return b;
}

sf::Vector2f maxVector(const sf::Vector2f& a, const sf::Vector2f& b)
{
	if (a.x > b.x && a.y > b.y)
		return a;
	else
		return b;
}

sf::Vector2f getRectangleForWrappedText(std::string text, int characterSize, const sf::Font* font, float width)
{
	int charactersPerLine = std::floor( width / static_cast<float>(characterSize / 1.5f) );
	sf::Vector2f Rectangle;

	if (text.size() < charactersPerLine)
	{
		Rectangle.x = text.size() * characterSize;
		Rectangle.y = characterSize * 1.72f;
		return Rectangle;
	}


}

void wrapAndLayoutText(std::vector<sf::Text>& outTextLines, std::string text, int characterSize, const sf::Font* font, const sf::Vector2f& boundary, const sf::Vector2f& position)
{
	sf::Text renderText;
	renderText.setFont(*font);
	renderText.setPosition(position);
	renderText.setCharacterSize(characterSize);

	int linesThatFit = std::floor( boundary.y / static_cast<float>(characterSize) );
	//1.5f is a magic number, it works well with our current font choice, ideally we should look up the size of glyphs within the font
	int charactersPerLine = std::floor( boundary.x / static_cast<float>(characterSize / 1.5f) );

	if (text.size() < charactersPerLine)
	{
		renderText.setString( text );
		renderText.setPosition(position.x, position.y);
		outTextLines.push_back(renderText);
		return;
	}

	std::size_t previousCutoffPoint = 0;
	std::size_t cutoffPoint = 0;
	int i = 0;
	while (cutoffPoint < text.size())
	{

		if (previousCutoffPoint == 0)
		{
			std::size_t newlinePos = text.find_first_of('\n');
			cutoffPoint = text.rfind(' ', charactersPerLine);
			std::size_t reverseNewlinePos = text.rfind('\n', charactersPerLine);
			if (cutoffPoint == std::string::npos || reverseNewlinePos > cutoffPoint)
			{
				if (reverseNewlinePos != std::string::npos)
					cutoffPoint = reverseNewlinePos;
			}
			
			//cutoffPoint = m_text.rfind(' ', charactersPerLine);
			if (cutoffPoint == std::string::npos)
				break;
			//if (m_text.rfind('\n', charactersPerLine) < cutoffPoint)
			if (newlinePos <= cutoffPoint)
			{
				cutoffPoint = newlinePos;
				//cutoffPoint = m_text.rfind('\n', charactersPerLine);
			} 
			text.erase(cutoffPoint, 1);

		} else
		{
			std::size_t newlinePos = text.find_first_of('\n', previousCutoffPoint + charactersPerLine);
			cutoffPoint = text.rfind(' ', previousCutoffPoint + charactersPerLine);
			

			if (newlinePos < cutoffPoint)
			{
				cutoffPoint = newlinePos;
			} else if (previousCutoffPoint + charactersPerLine >= text.size())
			{
				cutoffPoint = std::max( text.rfind('\0', previousCutoffPoint + charactersPerLine), cutoffPoint );
			}

			//text.erase(cutoffPoint, 1);
		}

		if (cutoffPoint == std::string::npos)
			cutoffPoint = std::min(text.size(), static_cast<std::size_t>(previousCutoffPoint + charactersPerLine));
		
		if (cutoffPoint > text.size())
			break;
		std::string temp = text.substr( previousCutoffPoint, cutoffPoint - previousCutoffPoint);
		int firstNonWhitespace = temp.find_first_not_of(' ');
		temp = temp.substr(firstNonWhitespace, temp.size() - firstNonWhitespace);
		renderText.setString( temp );
		
		renderText.setPosition(position.x, position.y + font->getLineSpacing(characterSize) * i);
		outTextLines.push_back(renderText);
		previousCutoffPoint = cutoffPoint;
		++i;
		if (i > linesThatFit)
			break;
	}
}


/*##########################################################################################################################################################
Entity
*/

Entity::Entity() : m_isManagedByLayout(false), m_fillWidth(0), m_fillHeight(0), m_isVisible(true)
{
	
}

void Entity::setDefaultSize(const sf::Vector2f& size)
{
	m_defaultSize = size;
}

void Entity::setMinimumSize(const sf::Vector2f& size)
{
	m_minimumSize = size;
}

void Entity::setMaximumSize(const sf::Vector2f& size)
{
	m_maximumSize = size;
}

void Entity::setFillWidth(int i)
{
	m_fillWidth = i;
}

void Entity::setFillHeight(int i)
{
	m_fillHeight = i;
}

const sf::Vector2f& Entity::getDefaultSize() const
{
	return m_defaultSize;
}

const sf::Vector2f& Entity::getMinimumSize() const
{
	return m_minimumSize;
}

const sf::Vector2f& Entity::getMaximumSize() const
{
	return m_maximumSize;
}

int Entity::getFillWidth() const
{
	return m_fillWidth;
}

int Entity::getFillHeight() const
{
	return m_fillHeight;
}

const sf::Vector2f& Entity::getSize() const
{
	return m_displaySize;
}

void Entity::setMargins(float left, float top, float right, float bottom)
{
	m_margins = Margins(left, right, top, bottom);
}

void Entity::setMargins(const Margins& Margins)
{
	m_margins = Margins;
}

const Entity::Margins& Entity::getMargins() const
{
	return m_margins;
}

/*##########################################################################################################################################################
Implementation: Panel
*/

Panel::Panel(Layout* layout) : m_useTexture(false), m_layout(layout)
{
	m_rectShape.setFillColor(Default::PanelBackground);
}

void Panel::setBackgroundColor(const sf::Color& color)
{
	m_rectShape.setFillColor(color);
}

void Panel::addEntity(std::shared_ptr<Entity> entity)
{
	m_entities.push_back(entity);
	m_layout->add(entity);
	//entity->onLayout(m_rectShape.getPosition(), m_rectShape.getSize());
	//TODO: change to getTotalMinSize()
	if (m_layout->getSize().x > m_minimumSize.x && m_minimumSize.x != 0)
		m_minimumSize.x = m_layout->getSize().x;
	if (m_layout->getSize().y > m_minimumSize.y && m_minimumSize.y != 0)
		m_minimumSize.y = m_layout->getSize().y;
	/*if (m_minimumSize.x != 0)
	{
		if (m_layout->getLayoutOrientation() == Layout::HORIZONTAL)
			m_minimumSize.x = m_layout->m_totalMinSize.x;
		else
			m_minimumSize.x = std::max(m_minimumSize.x, entity->getMinimumSize().x);
	}
	if (m_minimumSize.y != 0)
	{
		if (m_layout->getLayoutOrientation() == Layout::VERTICAL)
			m_minimumSize.y = m_layout->m_totalMinSize.y;
		else
			m_minimumSize.y = std::max(m_minimumSize.y, entity->getMinimumSize().y);
	}*/
}

void Panel::onLayout(const sf::Vector2f& position, const sf::Vector2f& size)
{
	m_isManagedByLayout = true;
	m_rectShape.setPosition(position);
	
	//m_minimumSize = size;
	
	m_displaySize = size;

	if (m_layout->getSize().x > m_displaySize.x)
		m_displaySize.x = m_layout->getSize().x;
	if (m_layout->getSize().y > m_displaySize.y)
		m_displaySize.y = m_layout->getSize().y;

	m_rectShape.setSize(m_displaySize);
}

void Panel::onDraw(sf::RenderTarget* renderTarget)
{
	sf::Color test = m_rectShape.getFillColor();
	renderTarget->draw(m_rectShape);
	for (int i = 0; i < m_entities.size(); ++i)
	{
		if (m_entities[i]->getIsVisible())
			m_entities[i]->onDraw(renderTarget);
	}
}

bool Panel::onEvent(const sf::Event& ev)
{
	bool handledEvent = false;

	switch (ev.type)
	{
		case sf::Event::Resized:
			//recalculate the correct size of panel
			if (!m_isManagedByLayout)
				m_rectShape.setSize(sf::Vector2f(ev.size.width, ev.size.height));

			m_layout->onResize(m_rectShape.getPosition(), m_rectShape.getSize());
			m_rectShape.setSize(sf::Vector2f(std::max(m_rectShape.getSize().x, m_layout->getSize().x), std::max(m_rectShape.getSize().y, m_layout->getSize().y)));
			break;
	}

	for (int i = 0; i < m_entities.size(); ++i)
	{
		if (m_entities[i]->onEvent(ev))
			handledEvent = true;
	}

	return handledEvent;
}


/*##########################################################################################################################################################
Implementation: Layout
*/

void HorizontalBoxLayout::add(std::shared_ptr<Entity> e)
{
	adjustMinimumFittingRectangle(e);
	Layout::add(e);
}

void HorizontalBoxLayout::onResize(const sf::Vector2f& startPosition, const sf::Vector2f& allottedSize)
{
	//adjust all member entities according to a horizontal box layout

	float totalFill = 0;
	m_minimumFittingRectangle.x = 0;
	m_minimumFittingRectangle.y = 0;

	for (int i = 0; i < m_entities.size(); ++i)
	{
		if (m_entities[i]->getIsVisible())
		{
			totalFill += m_entities[i]->getFillWidth();
			adjustMinimumFittingRectangle(m_entities[i]);
		}
	}


	float xIncrement = 0;//m_entities.size() > 0 ? m_entities[0]->getMargins().left : 0;
	for (int i = 0; i < m_entities.size(); ++i)
	{
		if (m_entities[i]->getIsVisible())
		{
			/*sf::Vector2f childSize;
			childSize.x = std::min(m_entities[i]->getSize().x, allottedSize.x);
			if (childSize.x == 0)
				childSize.x = m_entities[i]->getSize().x;
			childSize.y = std::min(m_entities[i]->getSize().y, allottedSize.y);
			if (childSize.y == 0)
				childSize.y = m_entities[i]->getSize().y;*/
			sf::Vector2f childSize = minVector(m_entities[i]->getDefaultSize(), allottedSize);
			bool fillHeight = false;

			if (childSize.y == 0 && m_entities[i]->getMinimumSize().y == 0 || m_entities[i]->getFillHeight() > 0)
			{
				childSize.y = allottedSize.y;
				fillHeight = true;
			} else if (m_entities[i]->getMinimumSize().y > 0)
			{
				childSize.y = m_entities[i]->getMinimumSize().y;
			}

			if (m_entities[i]->getMinimumSize().x > 0)
				childSize.x = std::max(childSize.x, std::min(m_entities[i]->getMinimumSize().x, allottedSize.x));

			/*
			Note: this is not correct maximum size handling. I will still leave it here for reference, until a later time when I get back to this.
			For now, treat maximum size as broken if we use fill width.
			Since this is a work in progress, no similar code can be found in the vertical layout yet. So they will have different behaviour when it comes
			to filling and maximum size.
			*/
			if (m_entities[i]->getFillWidth() > 0)
			{
				/*if (m_entities[i]->getMaximumSize().x > 0)
					childSize.x = std::min(m_entities[i]->getMaximumSize().x, (m_entities[i]->getFillWidth() / totalFill) * (allottedSize.x - m_totalMinSize.x + m_entities[i]->getMinimumSize().x - m_entities[i]->getMargins().right) );
				else*/
				childSize.x = (m_entities[i]->getFillWidth() / totalFill) * (allottedSize.x - m_minimumFittingRectangle.x + m_entities[i]->getMinimumSize().x - m_entities[i]->getMargins().right - m_entities[i]->getMargins().left);
			}

			if (fillHeight && m_entities[i]->getMaximumSize().y > 0)
				childSize.y = std::min(m_entities[i]->getMaximumSize().y, childSize.y);


			if (fillHeight)
			{
				m_entities[i]->onLayout( sf::Vector2f(startPosition.x + xIncrement + m_entities[i]->getMargins().left, startPosition.y + m_entities[i]->getMargins().top),
									sf::Vector2f(std::max(m_entities[i]->getMinimumSize().x, childSize.x - m_entities[i]->getMargins().left - m_entities[i]->getMargins().right),
												std::max(m_entities[i]->getMinimumSize().y, childSize.y - m_entities[i]->getMargins().top - m_entities[i]->getMargins().bottom)));
				//NOTE: -2 seems like a hack to adjust for the not intended spacing between widgets
				if (m_entities[i]->getMinimumSize().x > (childSize.x - m_entities[i]->getMargins().left - m_entities[i]->getMargins().right))
					xIncrement += m_entities[i]->getMargins().left + m_entities[i]->getMargins().right;
			} else
			{
				m_entities[i]->onLayout( sf::Vector2f(startPosition.x + xIncrement + m_entities[i]->getMargins().left, startPosition.y + m_entities[i]->getMargins().top),
									sf::Vector2f(childSize.x, childSize.y));
			}
			xIncrement += childSize.x;
			/*if (m_entities[i]->getMinimumSize().x == 0)
				xIncrement += m_entities[i]->getSize().x;*/
		}
	}
}

void HorizontalBoxLayout::adjustMinimumFittingRectangle(const std::shared_ptr<Entity>& e)
{
	m_minimumFittingRectangle.x += e->getMinimumSize().x - e->getMargins().left - e->getMargins().right;
	m_minimumFittingRectangle.y = std::max(m_minimumFittingRectangle.y, e->getMinimumSize().y + e->getMargins().top + e->getMargins().bottom);
}



void VerticalBoxLayout::add(std::shared_ptr<Entity> e)
{
	adjustMinimumFittingRectangle(e);
	Layout::add(e);
}

void VerticalBoxLayout::onResize(const sf::Vector2f& startPosition, const sf::Vector2f& allottedSize)
{
	//adjust all member entities according to a vertical box layout
	
	float totalFill = 0;
	m_minimumFittingRectangle.x = 0;
	m_minimumFittingRectangle.y = 0;
	float totalInvisibleMinSize = 0;

	for (int i = 0; i < m_entities.size(); ++i)
	{
		if (m_entities[i]->getIsVisible())
		{
			//this should probably change later if we start allowing relative positional changes inside layouts for the children
			totalFill += m_entities[i]->getFillHeight();
			adjustMinimumFittingRectangle(m_entities[i]);
		}
	}

	float yIncrement = 0;
	for (int i = 0; i < m_entities.size(); ++i)
	{
		if (m_entities[i]->getIsVisible())
		{
			sf::Vector2f childSize = minVector(m_entities[i]->getDefaultSize(), allottedSize);
			bool fillWidth = false;

			if (childSize.x == 0 && m_entities[i]->getMinimumSize().x == 0 || m_entities[i]->getFillWidth() > 0)
			{
				childSize.x = allottedSize.x;
				fillWidth = true;
			} else if (m_entities[i]->getMinimumSize().x > 0)
			{
				childSize.x = m_entities[i]->getMinimumSize().x;
			}

			if (m_entities[i]->getMinimumSize().y > 0)
				childSize.y = std::max(childSize.y, std::min(m_entities[i]->getMinimumSize().y, allottedSize.y));
			//solving max height like this means using fillheight will override maximum height, but that could be fine
			if (m_entities[i]->getMaximumSize().y > 0)
				childSize.y = std::min(childSize.y, m_entities[i]->getMaximumSize().y);
			if (m_entities[i]->getFillHeight() > 0)
				childSize.y = (m_entities[i]->getFillHeight() / totalFill) * (allottedSize.y - m_minimumFittingRectangle.y + m_entities[i]->getMinimumSize().y - m_entities[i]->getMargins().bottom - m_entities[i]->getMargins().top);

			if (fillWidth && m_entities[i]->getMaximumSize().x > 0)
				childSize.x = std::min(m_entities[i]->getMaximumSize().x, childSize.x);

			if (fillWidth)
			{
				m_entities[i]->onLayout( sf::Vector2f(startPosition.x + m_entities[i]->getMargins().left, startPosition.y + yIncrement + m_entities[i]->getMargins().top),
									sf::Vector2f(std::max(m_entities[i]->getMinimumSize().x, childSize.x - m_entities[i]->getMargins().left - m_entities[i]->getMargins().right),
												std::max(m_entities[i]->getMinimumSize().y, childSize.y - m_entities[i]->getMargins().top - m_entities[i]->getMargins().bottom)));

				if (m_entities[i]->getMinimumSize().y >= childSize.y - m_entities[i]->getMargins().top - m_entities[i]->getMargins().bottom)
					yIncrement += m_entities[i]->getMargins().bottom + m_entities[i]->getMargins().top;


			} else
			{
				m_entities[i]->onLayout( sf::Vector2f(startPosition.x + m_entities[i]->getMargins().left, startPosition.y + yIncrement + m_entities[i]->getMargins().top),
										sf::Vector2f(childSize.x, childSize.y));
			}

			
			yIncrement += childSize.y;
		}
	}
}

void VerticalBoxLayout::adjustMinimumFittingRectangle(const std::shared_ptr<Entity>& e)
{
	m_minimumFittingRectangle.x = std::max(m_minimumFittingRectangle.x, e->getMinimumSize().x + e->getMargins().left + e->getMargins().right);
	m_minimumFittingRectangle.y += e->getMinimumSize().y - e->getMargins().bottom - e->getMargins().top;
}

/*##########################################################################################################################################################
Concrete entities
*/



class DisassemblyOutput : public Entity
{
	public:
		virtual void onLayout( const sf::Vector2f& position, const sf::Vector2f& size)
		{

		}

		virtual void onDraw(sf::RenderTarget* window)
		{

		}

		virtual bool onEvent(const sf::Event& ev)
		{

		}
	private:
};


/*##########################################################################################################################################################
Implementation: TextureRenderWidget
*/

TextureRenderWidget::TextureRenderWidget() : m_renderTexture(0)
{}

TextureRenderWidget::TextureRenderWidget(const sf::Texture* texture)
{
	setTexture(texture);
}

void TextureRenderWidget::onLayout(const sf::Vector2f& position, const sf::Vector2f& size)
{
	m_sprite.setPosition(position);
	//m_sprite.setTextureRect( sf::IntRect(sf::Vector2i(0, 0), sf::Vector2i(size.x, size.y)) );
	sf::Vector2u textureSize = m_renderTexture->getSize();
	m_sprite.setScale(size.x / static_cast<float>(textureSize.x), size.y / static_cast<float>(textureSize.y));
	m_displaySize = size;
	
}

void TextureRenderWidget::onDraw(sf::RenderTarget* renderTarget)
{
	renderTarget->draw(m_sprite);
}

bool TextureRenderWidget::onEvent(const sf::Event& ev)
{
	return false;
}

void TextureRenderWidget::setTexture(const sf::Texture* texture)
{
	m_renderTexture = texture;
	m_minimumSize.x = texture->getSize().x;
	m_minimumSize.y = texture->getSize().y;
	m_sprite.setTexture(*m_renderTexture);
}

/*##########################################################################################################################################################
Implementation: Button
*/

Button::Button()
{
	init();
}

Button::Button(const std::string& text)
{
	init();
	m_text.setString(text);
	
	if (m_defaultSize.x == 0 && m_defaultSize.y == 0 && m_minimumSize.x == 0 && m_minimumSize.y == 0)
	{
		m_minimumSize.x = m_text.getLocalBounds().width + 10;
		if (m_maximumSize.x < m_minimumSize.x && m_maximumSize.x > 0)
			m_maximumSize.x = std::max(m_minimumSize.x, m_maximumSize.x);
		m_minimumSize.y = m_text.getLocalBounds().height * 1.72 + 10; //m_text.getCharacterSize();
		if (m_maximumSize.y < m_minimumSize.y && m_maximumSize.y > 0)
			m_maximumSize.y = std::max(m_minimumSize.y, m_maximumSize.y);
	}
}

void Button::init()
{
	m_hasImage = false;
	m_text.setFont(*Default::TextFont);
	m_text.setCharacterSize(12);
	m_imageAlignment = ImageAlignment::TOP;
	m_buttonState = STATE_UP;
	m_hover = false;

	//default background is transparent
	m_defaultColor = sf::Color(0,0,0,0);
	m_downColor = Default::ButtonDown;
	m_hoverColor = Default::ButtonHover;
}

void Button::setImage(const std::shared_ptr<sf::Texture> image)
{
	m_buttonImage = image;
	m_buttonImageSprite.setTexture(*m_buttonImage.get());

	switch (m_imageAlignment)
	{
		case ImageAlignment::TOP:
		{
			float comparisonWidth = std::max(m_text.getLocalBounds().width, static_cast<float>(m_buttonImage->getSize().x + 10));
			if (m_minimumSize.x < comparisonWidth)
			{
				m_minimumSize.x = comparisonWidth;
				if (m_maximumSize.x < m_minimumSize.x && m_maximumSize.x > 0)
					m_maximumSize.x = std::max(m_minimumSize.x, m_maximumSize.x);
			}
			if (m_minimumSize.y < (m_text.getLocalBounds().height * 1.72 + m_buttonImage->getSize().y))
			{
				m_minimumSize.y = (m_text.getLocalBounds().height * 1.72 + m_buttonImage->getSize().y);
				if (m_maximumSize.y < m_minimumSize.y && m_maximumSize.y > 0)
					m_maximumSize.y = std::max(m_minimumSize.y, m_maximumSize.y);
			}
			break;
		}
	}
	
	m_hasImage = true;
}

void Button::setTextColor(const sf::Color& color)
{
	m_text.setColor(color);
}

void Button::setDefaultColor(const sf::Color& color)
{
	m_defaultColor = color;
}

void Button::setDownColor(const sf::Color& color)
{
	m_downColor = color;
}

void Button::setHoverColor(const sf::Color& color)
{
	m_hoverColor = color;
}

void Button::setOnPressedCallback(std::function<void()> function)
{
	m_onPressed = function;
}

void Button::onLayout(const sf::Vector2f& position, const sf::Vector2f& size)
{
	m_position = position;
	m_displaySize = size;
}

void Button::onDraw(sf::RenderTarget* renderTarget)
{
	float textX = 0;
	float textY = 0;
	float imageX = 0;
	float imageY = 0;

	if (m_hover && m_buttonState != ButtonState::STATE_DOWN)
	{
		sf::RectangleShape hoverRect(sf::Vector2f(m_displaySize.x - 2, m_displaySize.y - 2));
		hoverRect.setFillColor(m_hoverColor);
		hoverRect.setPosition(m_position);
		renderTarget->draw(hoverRect);
	} else if (m_buttonState != ButtonState::STATE_DOWN)
	{
		//we could do this small optimization on all drawing states, but the most common scenario ought to be full transparency on the button up state only
		if (m_defaultColor.a > 0)
		{
			sf::RectangleShape backgroundRect(sf::Vector2f(m_displaySize.x - 2, m_displaySize.y - 2));
			backgroundRect.setFillColor(m_defaultColor);
			backgroundRect.setPosition(m_position);
			renderTarget->draw(backgroundRect);
		}
	} else if (m_buttonState == STATE_DOWN)
	{
		sf::RectangleShape downRect(sf::Vector2f(m_displaySize.x - 2, m_displaySize.y - 2));
		downRect.setFillColor(m_downColor);
		downRect.setPosition(m_position);
		renderTarget->draw(downRect);
		textX += 1;
		textY += 1;
		imageX += 1;
		imageY += 1;
	}
	
	if (m_hasImage)
	{
		//TODO: add the 3 other alignment modes
		switch (m_imageAlignment)
		{
			case ImageAlignment::TOP:
				//m_buttonImageSprite.setPosition;
				float totalWidth = (m_buttonImage->getSize().x + m_text.getLocalBounds().width);
				imageX += m_position.x + (m_displaySize.x / 2.f) - (m_buttonImage->getSize().x / 2.f);
				float totalHeight = (m_buttonImage->getSize().y + m_text.getLocalBounds().height * 1.72f);
				imageY += m_position.y + (m_displaySize.y / 2.f) - (m_buttonImage->getSize().y / 2.f) - ((m_text.getLocalBounds().height * 1.72f) / 2.f);
				m_buttonImageSprite.setPosition(imageX, imageY);
				renderTarget->draw(m_buttonImageSprite);
				textX += m_position.x + (m_displaySize.x / 2.f) - (m_text.getLocalBounds().width / 2.f);
				textY += imageY + m_buttonImage->getSize().x;
				break;
		}
	} else
	{
		//if we don't have an image, center the text in the button
		textX += m_position.x + (m_displaySize.x / 2.f) - (m_text.getLocalBounds().width / 2.f);
		textY += m_position.y + (m_displaySize.y / 2.f) - ((m_text.getLocalBounds().height * 1.72f) / 2.f);
	}
	
	m_text.setPosition(textX, textY);
	
	renderTarget->draw(m_text);
}

bool Button::onEvent(const sf::Event& ev)
{
	switch (ev.type)
	{
		case sf::Event::MouseMoved:
			if (isInBoundary(ev.mouseMove.x, ev.mouseMove.y))
			{
				m_hover = true;
				return true;
			} else
			{
				m_hover = false;
			}
			break;
		case sf::Event::MouseButtonPressed:
			if (isInBoundary(ev.mouseButton.x, ev.mouseButton.y))
			{
				if (ev.mouseButton.button == sf::Mouse::Button::Left)
					m_buttonState = STATE_DOWN;
				return true;
			}
			break;
		case sf::Event::MouseButtonReleased:
			if (ev.mouseButton.button == sf::Mouse::Button::Left)
			{
				m_buttonState = STATE_UP;
				if (isInBoundary(ev.mouseButton.x, ev.mouseButton.y))
					if (m_onPressed)
						m_onPressed();
			}
			return true;
			break;
		case sf::Event::LostFocus:
			m_buttonState = STATE_UP;
			break;
	}
	

	return false;
}

bool Button::isInBoundary(int x, int y)
{
	if (( (x > m_position.x) && (x < (m_position.x + m_displaySize.x)) ) && ( (y > m_position.y) && (y < (m_position.y + m_displaySize.y)) ))
		return true;
	
	return false;
}

/*##########################################################################################################################################################
Implementation: ToggleButton
*/

ToggleButton::ToggleButton(const std::string& text) : Button(text), m_currentButtonDown(false), m_toggleMode(CLICK_TOGGLE)
{

}

bool ToggleButton::onEvent(const sf::Event& ev)
{
	switch (ev.type)
	{
		case sf::Event::MouseMoved:
			if (isInBoundary(ev.mouseMove.x, ev.mouseMove.y))
			{
				m_hover = true;
				return true;
			} else
			{
				m_hover = false;
			}
			break;
		case sf::Event::MouseButtonPressed:
			if (isInBoundary(ev.mouseButton.x, ev.mouseButton.y))
			{
				if (ev.mouseButton.button == sf::Mouse::Button::Left)
				{
					if (m_buttonState != STATE_DOWN)
						m_currentButtonDown = true;
					if (m_toggleMode == CLICK_SET)
						m_buttonState = STATE_DOWN;
					else
					{
						if (m_buttonState == STATE_DOWN)
							m_buttonState = STATE_UP;
						else
							m_buttonState = STATE_DOWN;
					}
				}
				return true;
			}
			break;
		case sf::Event::MouseButtonReleased:
		{
			if (m_buttonState == STATE_DOWN)
			{
				ButtonState oldState = m_buttonState;
				Button::onEvent(ev);
				m_buttonState = oldState;
			}

			if (m_currentButtonDown && !isInBoundary(ev.mouseButton.x, ev.mouseButton.y))
			{
				m_buttonState = STATE_UP;
			}
			m_currentButtonDown = false;
			return true;
			
			break;
		}
	}
	

	return false;
}

void ToggleButton::setMode(ToggleMode mode)
{
	m_toggleMode = mode;
}

void ToggleButton::setChecked(bool state)
{
	if (state)
		m_buttonState = STATE_DOWN;
	else
		m_buttonState = STATE_UP;
}

bool ToggleButton::getIsChecked()
{
	if (m_buttonState == STATE_DOWN)
		return true;
	else
		return false;
}

/*##########################################################################################################################################################
Implementation: TabbedInterface
*/

TabbedInterface::TabbedInterface() : m_activeTab(-1)
{
	m_headerColor = Default::TabHeader;
	m_headerHoverColor = Default::TabHeaderHover;
	m_backgroundColor = Default::TabHeaderBackground;

	m_tabbedInterfacePanel = std::make_shared<Panel>(new VerticalBoxLayout());
	m_tabbedInterfacePanel->setFillWidth(1);
	m_tabbedInterfacePanel->setFillHeight(1);
	//m_tabbedInterfacePanel->setMargins(Margins(5));
	m_tabbedInterfacePanel->setBackgroundColor(sf::Color::Transparent);

		m_tabHeaderPanel = std::make_shared<Panel>(new HorizontalBoxLayout());
		m_tabHeaderPanel->setFillWidth(1);
		m_tabHeaderPanel->setMinimumSize(sf::Vector2f(0,1));
		m_tabHeaderPanel->setMargins(5,0,5,0);
		m_tabHeaderPanel->setBackgroundColor(sf::Color::Transparent);
		//m_tabHeaderPanel->setBackgroundColor(sf::Color(0,0,0,0));
		m_tabbedInterfacePanel->addEntity(m_tabHeaderPanel);

		m_tabContentPanel = std::make_shared<Panel>(new HorizontalBoxLayout());
		m_tabContentPanel->setFillHeight(1);
		m_tabContentPanel->setMargins(2,0,2,2);
		m_tabContentPanel->setBackgroundColor(sf::Color::Transparent);
		m_tabbedInterfacePanel->addEntity(m_tabContentPanel);
}

void TabbedInterface::addPanel(const std::shared_ptr<Panel>& panel, const std::string& text)
{
	TabPanel newTab;
	int newTabIndex = m_tabPanels.size();
	newTab.button = std::make_shared<ToggleButton>(text);
	newTab.button->setMode(ToggleButton::CLICK_SET);
	newTab.button->setMargins(0, 2, 0, -2);
	newTab.button->setDefaultColor(m_backgroundColor);
	newTab.button->setDownColor(m_headerColor);
	newTab.button->setHoverColor(m_headerHoverColor);
	newTab.button->setOnPressedCallback([newTabIndex, this]() 
	{
		selectTab(newTabIndex);
	}
	);
	newTab.panel = panel;
	panel->setBackgroundColor(m_headerColor);
	panel->setFillHeight(1);
	panel->setFillWidth(1);
	
	m_tabPanels.push_back(newTab);
	m_tabHeaderPanel->addEntity(newTab.button);
	m_tabContentPanel->addEntity(newTab.panel);
	
	if (m_activeTab == -1)
		m_activeTab = newTabIndex;
	selectTab(newTabIndex);
}

void TabbedInterface::selectTab(int tabIndex)
{
	m_tabPanels[m_activeTab].panel->setIsVisible(false);
	m_tabPanels[m_activeTab].button->setChecked(false);
	m_tabPanels[tabIndex].panel->setIsVisible(true);
	m_tabPanels[tabIndex].button->setChecked(true);

	sf::Event ev;
	ev.type = sf::Event::Resized;
	ev.size.width = m_displaySize.x;
	ev.size.height = m_displaySize.y;
	m_tabbedInterfacePanel->onEvent(ev);
	m_activeTab = tabIndex;
}

void TabbedInterface::setHeaderColor(const sf::Color& color)
{
	m_headerColor = color;
}

void TabbedInterface::setHeaderHoverColor(const sf::Color& color)
{
	m_headerHoverColor = color;
}

void TabbedInterface::setBackgroundColor(const sf::Color& color)
{
	m_backgroundColor = color;
}

void TabbedInterface::onLayout(const sf::Vector2f& position, const sf::Vector2f& size)
{
	m_tabbedInterfacePanel->onLayout(position, size);
	m_position = position;
	m_displaySize = size;
}

void TabbedInterface::onDraw(sf::RenderTarget* renderTarget)
{
	sf::RectangleShape background(m_displaySize);
	background.setPosition(m_position);
	background.setFillColor(m_backgroundColor);
	renderTarget->draw(background);
	for (int i = 0; i < m_tabPanels.size(); ++i)
	{
		if (m_tabPanels[i].panel->getIsVisible())
			m_tabPanels[i].panel->onDraw(renderTarget);
		m_tabPanels[i].button->onDraw(renderTarget);
	}
}

bool TabbedInterface::onEvent(const sf::Event& ev)
{
	
	m_tabbedInterfacePanel->onEvent(ev);
	return true;
}

/*##########################################################################################################################################################
Implementation: Label
*/

Label::Label()
{
	init();
}

Label::Label(const char* text) : m_text(text)
{
	init();
}

Label::Label(TextDelegate textDelegate) : m_textDelegate(textDelegate)
{
	init();
}

void Label::init()
{
	m_characterSize = 12;
	m_borderWidth = 1;
	m_borderStyle = BorderStyle::None;
	m_borderColor = sf::Color::Black;
	m_backgroundColor = sf::Color(0,0,0,90);

	m_minimumSize.x = m_characterSize * std::min(m_text.find_first_of('\n'), m_text.size());
	m_minimumSize.y = m_characterSize*1.72;
}

void Label::layoutText()
{
	if (m_text.empty())
		return;

	m_renderTextLines.clear();
	
	getRectangleForWrappedText(m_text, m_characterSize, Default::TextFont, m_displaySize.x);
	wrapAndLayoutText(m_renderTextLines, m_text, m_characterSize, Default::TextFont, m_displaySize, sf::Vector2f(m_position.x + m_borderWidth, m_position.y + m_borderWidth));
}

void Label::setBorderWidth(int width)
{
	m_borderWidth = width;
}

void Label::setBorderStyle(BorderStyle bs)
{
	m_borderStyle = bs;
}

void Label::setBorderColor(const sf::Color& color)
{
	m_borderColor = color;
}

void Label::setBackgroundColor(const sf::Color& color)
{
	m_backgroundColor = color;
}

void Label::setCharacterSize(int size)
{
	m_characterSize = size;
}

void Label::onLayout(const sf::Vector2f& position, const sf::Vector2f& size)
{
	m_position = position;
	m_displaySize = size;

	layoutText();

	if (m_textDelegate)
		m_text = m_textDelegate();
}

void Label::onDraw(sf::RenderTarget* renderTarget)
{
	if (m_borderStyle == BorderStyle::Line)
	{
		sf::RectangleShape border(sf::Vector2f(m_displaySize.x, m_displaySize.y));
		border.setFillColor(m_backgroundColor);
		border.setOutlineColor(m_borderColor);
		border.setOutlineThickness(m_borderWidth);
		border.setPosition(m_position);
		renderTarget->draw(border);
	}

	if (m_textDelegate)
	{
		m_text = m_textDelegate();
		layoutText();
	}

	for (auto iter = m_renderTextLines.begin(); iter != m_renderTextLines.end(); ++iter)
	{
		//iter->setPosition(m_position.x, m_position.y * m_characterSize);
		renderTarget->draw(*iter);
	}
}

bool Label::onEvent(const sf::Event& ev)
{
	return false;
}

/*##########################################################################################################################################################
Implementation: ListView
*/


ListView::ListView() : m_itemCallback(nullptr)
{

}

void ListView::setItemCallback(IItemCallback* iic)
{
	m_itemCallback = iic;
}

void ListView::onLayout(const sf::Vector2f& position, const sf::Vector2f& size)
{
	m_position = position;
	m_displaySize = size;
}

void ListView::onDraw(sf::RenderTarget* renderTarget)
{
	

	//draw encompassing frame for the listview
	//sf::ConvexShape
	sf::RectangleShape background(m_displaySize);
	background.setPosition(m_position);
	background.setFillColor(Default::WidgetBackground);
	renderTarget->draw(background);

	if (m_itemCallback == nullptr)
		return;
	sf::Vector2f position = m_position;
	position.x += 2;
	sf::Vector2f size = m_displaySize;
	size.x -= 4;

	if (m_itemCallback != nullptr)
	{
		float accumHeight = 0;
		for (int i = 0; i < m_itemCallback->getNumItems() && (accumHeight + m_itemCallback->getSize(i).y) < size.y; ++i)
		{
			m_itemCallback->draw(i, sf::Vector2f(position.x, position.y + accumHeight), size, renderTarget);
			accumHeight += m_itemCallback->getSize(i).y;
		}
	}
}

bool ListView::onEvent(const sf::Event& ev)
{


	return false;
}

void ListView::DefaultDrawItemText(const std::string&, const sf::Vector2f&, const sf::Vector2f&, sf::RenderTarget*)
{

}

} /* end namespace UI */
