/**
Very simple header only GUI library for SFML.

Create a "root" layout panel as your first component. Call onEvent() passing in your SFML event object to inject input and OS events to the GUI.
Call onDraw() in your drawing code on the root panel to ensure everything gets drawn.

To get dynamic resizing of the GUI, inject an sf::Event::Resized with the new size in your resizing event. This is



*/

#pragma once

#include "SFML/Graphics.hpp"
#include <vector>
#include <list>
#include <memory>
#include <functional>

namespace UI {

namespace Default {
	const sf::Color PanelBackground = sf::Color(100, 100, 100);
	const sf::Color WidgetBackground = sf::Color(80, 80, 80, 200);
	const sf::Color ButtonHover = sf::Color(120, 140, 120);
	const sf::Color ButtonDown = sf::Color(130, 150, 130);

	const sf::Color TabHeader = sf::Color(130,130,130);
	const sf::Color TabHeaderBackground = sf::Color(100,100,100);
	const sf::Color TabHeaderHover = sf::Color(110,110,110);

	extern sf::Font* TextFont;

	void Initialize(const std::string& fontFile);
}

enum Alignment {
	ALIGN_NONE, 
	ALIGN_LEFT, 
	ALIGN_RIGHT, 
	ALIGN_TOP,
	ALIGN_BOTTOM, 
	ALIGN_CENTER
};


void wrapAndLayoutText(std::vector<sf::Text>& outTextLines, std::string text, int characterSize, const sf::Font* font, const sf::Vector2f& boundary, const sf::Vector2f& position);

class Entity
{
	public:
		friend class Layout;

		Entity();
		
		/*
		initial size of an entity if nothing else is said, if it's >= minSize and <= maxSize
		
		*/
		void setDefaultSize(const sf::Vector2f& size);
		void setMinimumSize(const sf::Vector2f& size);
		void setMaximumSize(const sf::Vector2f& size);
		void setFillWidth(int);
		void setFillHeight(int);

		const sf::Vector2f& getDefaultSize() const;
		const sf::Vector2f& getMinimumSize() const;
		const sf::Vector2f& getMaximumSize() const;

		int getFillWidth() const;
		int getFillHeight() const;

		/*void setVAlign(Alignment alignment);
		void setHAlign(Alignment alignment);
		Alignment getAlignment() const;*/


		const sf::Vector2f& getSize() const;

		struct Margins
		{
			Margins() : left(0), right(0), top(0), bottom(0) {}
			Margins(float all)
			{
				left = right = top = bottom = all;
			}
			Margins(float inLeft, float inRight, float inTop, float inBottom) : left(inLeft), right(inRight), top(inTop), bottom(inBottom) {}

			float left;
			float right;
			float top;
			float bottom;
		};

		void setMargins(float left, float top, float right, float bottom);
		void setMargins(const Margins& Margins);
		const Margins& getMargins() const;

		void setIsVisible(bool state) { m_isVisible = state; }
		bool getIsVisible() const { return m_isVisible; }

		virtual void onLayout(const sf::Vector2f& position, const sf::Vector2f& size) = 0;

		virtual void onDraw(sf::RenderTarget*) = 0;
		virtual bool onEvent(const sf::Event&) = 0;

	protected:
		int m_fillWidth;
		int m_fillHeight;
		sf::Vector2f m_defaultSize;
		sf::Vector2f m_minimumSize;
		sf::Vector2f m_maximumSize;
		
		sf::Vector2f m_displaySize;

		Margins m_margins;

		bool m_isVisible;
		bool m_isManagedByLayout;
};

/**
Panel is a container entity. It is used to position other entities in a window and/or to take up space for layouts.
*/
class Panel : public Entity
{
	public:
		friend class Layout;

		Panel(class Layout* layout);

		void setBackgroundColor(const sf::Color& color);
		void addEntity(std::shared_ptr<Entity>);

		virtual void onLayout(const sf::Vector2f& position, const sf::Vector2f& size);
		virtual void onDraw(sf::RenderTarget*);
		virtual bool onEvent(const sf::Event&);

	private:
		void onResize(unsigned int, unsigned int);

		bool m_useTexture;
		sf::Texture m_texture;
		sf::RectangleShape m_rectShape;

		class Layout* m_layout;

		std::vector<std::shared_ptr<Entity> > m_entities;
};

typedef std::shared_ptr<Panel> PanelPtr;
/*##########################################################################################################################################################
Layout
*/
class Layout
{
	public:
		enum LayoutOrientation {
			NONE, VERTICAL, HORIZONTAL, BOTH
		};
		friend class Panel;

		virtual void add(std::shared_ptr<Entity> e)
		{
			e->m_isManagedByLayout = true;
			m_entities.push_back(e);
		}

		sf::Vector2f getSize() { return m_minimumFittingRectangle; }

		virtual void onResize(const sf::Vector2f& position, const sf::Vector2f& allottedSize) = 0;

		virtual int getLayoutOrientation() const { return NONE; }

	protected:
		virtual void adjustMinimumFittingRectangle(const std::shared_ptr<Entity>& e ) = 0;

		sf::Vector2f m_minimumFittingRectangle;
		std::vector<std::shared_ptr<Entity> > m_entities;
};

class AbsoluteLayout : public Layout
{
	public:
		virtual void onResize(const sf::Vector2f& position, const sf::Vector2f& allottedSize) {}

	protected:
		virtual void adjustMinimumFittingRectangle(const std::shared_ptr<Entity>& e ) {}
};

class HorizontalBoxLayout : public Layout
{
	public:
		virtual void add(std::shared_ptr<Entity> e);

		virtual void onResize(const sf::Vector2f& position, const sf::Vector2f&);
		virtual int getLayoutOrientation() const { return HORIZONTAL; }

	protected:
		virtual void adjustMinimumFittingRectangle(const std::shared_ptr<Entity>& e );
};

class VerticalBoxLayout : public Layout
{
	public:
		virtual void add(std::shared_ptr<Entity> e);

		virtual void onResize(const sf::Vector2f& position, const sf::Vector2f&);
		virtual int getLayoutOrientation() const { return VERTICAL; }

	protected:
		virtual void adjustMinimumFittingRectangle(const std::shared_ptr<Entity>& e );
};



/*##########################################################################################################################################################
TextureRenderWidget
*/

class TextureRenderWidget : public Entity
{
	public:
		TextureRenderWidget();
		TextureRenderWidget(const sf::Texture*);
		virtual void onLayout(const sf::Vector2f& position, const sf::Vector2f& size);

		virtual void onDraw(sf::RenderTarget*);
		virtual bool onEvent(const sf::Event&);

		void setTexture(const sf::Texture*);
	private:
		const sf::Texture*	m_renderTexture;
		sf::Sprite			m_sprite;
};

typedef std::shared_ptr<TextureRenderWidget> TextureRenderWidgetPtr;

enum class ImageAlignment : int {
	LEFT, RIGHT, TOP, BOTTOM
};

/*##########################################################################################################################################################
Button

This could be turned into a toggle button later, by inheriting from it, or by providing a boolean to switch it's behaviour into toggle mode rather than push button
BUG: if you hold down a button press, and then move mouse outside the window, it cannot detect the mouse release event, so it will get locked in a button down state until you get back to the window

TODO: extend support for images
*/
class Button : public Entity
{
	public:
		Button();

		Button(const std::string& text);

		void setImage(const std::shared_ptr<sf::Texture>);
		//image alignment relative to the text
		void setImageAlignment(ImageAlignment);

		void setTextColor(const sf::Color& color);
		void setDefaultColor(const sf::Color& color);
		void setDownColor(const sf::Color& color);
		void setHoverColor(const sf::Color& color);

		void setOnPressedCallback(std::function<void()>);

		virtual void onLayout(const sf::Vector2f& position, const sf::Vector2f& size);
		virtual void onDraw(sf::RenderTarget*);
		virtual bool onEvent(const sf::Event& ev);

		enum ButtonState {
			STATE_UP, STATE_DOWN
		};
	protected:
		sf::Vector2f m_position;
		ButtonState m_buttonState;

		bool m_hasImage;
		std::shared_ptr<sf::Texture> m_buttonImage;
		sf::Sprite m_buttonImageSprite;
		ImageAlignment m_imageAlignment;

		bool m_hover;

		sf::Text m_text;
		sf::Color m_defaultColor;
		sf::Color m_downColor;
		sf::Color m_hoverColor;

		bool isInBoundary(int x, int y);
	private:
		std::function<void()> m_onPressed;
		void init();
};

typedef std::shared_ptr<Button> ButtonPtr;

/*##########################################################################################################################################################
ToggleButton
*/

class ToggleButton : public Button
{
	public:
		ToggleButton(const std::string& text);

		virtual bool onEvent(const sf::Event& ev);

		enum ToggleMode {
			CLICK_SET, CLICK_TOGGLE
		};
		void setMode(ToggleMode);
		void setChecked(bool);
		bool getIsChecked();

	private:
		bool m_currentButtonDown;
		ToggleMode m_toggleMode;
};

typedef std::shared_ptr<ToggleButton> ToggleButtonPtr;

/*##########################################################################################################################################################
TabbedInterface
*/

class TabbedInterface : public Entity
{
	public:
		TabbedInterface();

		void addPanel(const std::shared_ptr<Panel>& panel, const std::string& text );

		void selectTab(int tabIndex);

		void setHeaderColor(const sf::Color&);
		void setHeaderHoverColor(const sf::Color&);
		void setBackgroundColor(const sf::Color&);

		virtual void onLayout(const sf::Vector2f& position, const sf::Vector2f& size);
		virtual void onDraw(sf::RenderTarget*);
		virtual bool onEvent(const sf::Event& ev);

	private:
		sf::Vector2f m_position;

		sf::Color m_headerColor;
		sf::Color m_headerHoverColor;
		sf::Color m_backgroundColor;

		struct TabPanel
		{
			std::shared_ptr<ToggleButton> button;
			std::shared_ptr<Panel> panel;
		};
		std::vector<TabPanel> m_tabPanels;

		std::shared_ptr<Panel> m_tabbedInterfacePanel;
		std::shared_ptr<Panel> m_tabHeaderPanel;
		std::shared_ptr<Panel> m_tabContentPanel;
		int m_activeTab;
};

typedef std::shared_ptr<TabbedInterface> TabbedInterfacePtr;

/*##########################################################################################################################################################
Label

Display arbritrary text, single or multiline. Wraps text at the end of a line, based on width.
Optional borders.

*/

enum class BorderStyle : int {
None, Line
};

class Label : public Entity
{
	public:
		typedef std::function<std::string()> TextDelegate;
		Label();
		Label(const char*);

		Label(TextDelegate textDelegate);

		void setBorderWidth(int width);
		
		void setBorderStyle(BorderStyle);

		/*
		Only visible if border style is != none
		*/
		void setBorderColor(const sf::Color&);

		/*
		Only visible if border style is != none
		*/
		void setBackgroundColor(const sf::Color&);

		void setCharacterSize(int size);

		virtual void onLayout(const sf::Vector2f& position, const sf::Vector2f& size);
		virtual void onDraw(sf::RenderTarget*);
		virtual bool onEvent(const sf::Event& ev);


	private:
		void init();
		void layoutText();

		sf::Vector2f m_position;

		TextDelegate m_textDelegate;
		std::string m_text;
		std::vector<sf::Text> m_renderTextLines;
		//sf::Text m_renderText;
		int m_characterSize;
		BorderStyle m_borderStyle;
		int m_borderWidth;
		sf::Color m_borderColor;
		sf::Color m_backgroundColor;
};

typedef std::shared_ptr<Label> LabelPtr;


/*##########################################################################################################################################################
ListView

Display a list of items in an organized fashion. We could make this support arbitrarily complex layouts (tiled large/small icons, detailed grid etc),
but lets stick to a regular list of items for now.

*/

class ListView : public Entity
{
	public:

		class IItemCallback
		{
			public:
				virtual int getNumItems() = 0;
				virtual sf::Vector2f getSize(int index) = 0;
				virtual void draw(int index, const sf::Vector2f& position, const sf::Vector2f& size, sf::RenderTarget*) = 0;
		};

		//typedef std::function<void(int, const sf::Vector2f&, const sf::Vector2f&, sf::RenderTarget*)> DrawItemDelegate;

		ListView();
		//ListView(int numItems);

		//void setNumItems(int);
		void setItemCallback(IItemCallback*);
		//void setOnDrawItemDelegate(DrawItemDelegate);

		virtual void onLayout(const sf::Vector2f& position, const sf::Vector2f& size);
		virtual void onDraw(sf::RenderTarget*);
		virtual bool onEvent(const sf::Event& ev);

		static void DefaultDrawItemText(const std::string&, const sf::Vector2f&, const sf::Vector2f&, sf::RenderTarget*);
	private:
		//DrawItemDelegate m_onDrawItemDelegate;
		IItemCallback* m_itemCallback;

		//sf::Color m_backgroundColor;
		//int m_numItems;
		sf::Vector2f m_position;
};

typedef std::shared_ptr<ListView> ListViewPtr;

} /* end namespace UI */
